# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
This module contains the mapping between gRPC calls and Python/dazl types.
"""

import asyncio
from datetime import datetime
from typing import (
    AbstractSet,
    Any,
    AsyncIterable,
    Collection,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Union,
    cast,
)
import warnings

from grpc import ChannelConnectivity
from grpc.aio import Channel, UnaryStreamCall

from .. import aio
from ... import LOG
from ..._gen.com.daml.ledger.api import v1 as lapipb
from ..._gen.com.daml.ledger.api.v1 import admin as lapiadminpb
from ...damlast.daml_lf_1 import PackageRef, TypeConName
from ...damlast.util import is_match
from ...prim import ContractData, ContractId, Party, datetime_to_timestamp, to_parties
from ...query import Filter, Queries, Query, parse_query
from .._offsets import END, End, LedgerOffsetRange, from_offset_until_forever
from .._retry import retry
from ..api_types import (
    ActAs,
    Admin,
    ArchiveEvent,
    Boundary,
    Command,
    CommandMeta,
    CreateEvent,
    ExerciseResponse,
    ParticipantMeteringReport,
    PartyInfo,
    ReadAs,
    Right,
    User,
    Version,
)
from ..config import Config
from ..config.access import TokenBasedAccessConfig
from ..errors import ProtocolWarning, _allow_cancel, _translate_exceptions
from .channel import create_channel
from .codec_aio import Codec

__all__ = ["Connection", "QueryStream"]


class Connection(aio.Connection):
    """
    An asynchronous (``asyncio``) connection to the Daml gRPC Ledger API.
    """

    def __init__(self, config: Config):
        self._config = config
        self._logger = config.logger
        self._channel = create_channel(config)
        self._codec = Codec(self, lookup=config.lookup)

    @property
    def config(self) -> Config:
        return self._config

    @property
    def channel(self) -> Channel:
        """
        Provides access to the underlying gRPC channel.
        """
        return self._channel

    @property
    def codec(self) -> Codec:
        return self._codec

    @property
    def is_closed(self) -> bool:
        return self._channel.get_state(try_to_connect=False) == ChannelConnectivity.SHUTDOWN

    async def __aenter__(self) -> "Connection":
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()

    async def open(self) -> None:
        """
        Does final validation of the token, including possibly fetching the ledger ID if it is not
        yet known.
        """
        if self._config.access.token_version == 2:
            # Daml 2.0 tokens do not contain party information, so an extra call to the server is
            # required in order to resolve our current set of rights
            if hasattr(self._config.access, "_set"):
                admin = False
                read_as = []  # type: List[Party]
                act_as = []  # type: List[Party]
                for right in await self.list_user_rights():
                    if right == Admin:
                        admin = True
                    elif isinstance(right, ReadAs):
                        read_as.append(right.party)
                    elif isinstance(right, ActAs):
                        act_as.append(right.party)

                # noinspection PyProtectedMember
                cast(TokenBasedAccessConfig, self._config.access)._set(
                    read_as=read_as, act_as=act_as, admin=admin
                )

        elif not self._config.access.ledger_id:
            # most calls to Daml 1.x ledgers require a ledger ID; if it wasn't supplied as part of
            # our token or we were never given a token in the first place, fetch the ledger ID from
            # the destination
            stub = lapipb.LedgerIdentityServiceStub(self._channel)
            response = await stub.GetLedgerIdentity(lapipb.GetLedgerIdentityRequest())

            self._logger.info("Connected to gRPC Ledger API, ledger ID: %s", response.ledger_id)
            self._config.access.ledger_id = response.ledger_id

    async def close(self) -> None:
        """
        Close the underlying channel. Once the channel is closed, future command submissions,
        streams in progress, and any future streams will fail.
        """
        await self._channel.close()

    # region Write API

    async def submit(
        self,
        __commands: Union[Command, Sequence[Command]],
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
        read_as: Union[None, Party, Collection[Party]] = None,
        act_as: Union[None, Party, Collection[Party]] = None,
    ) -> "None":
        """
        Submit one or more commands to the Ledger API.

        You should generally prefer trying to use :meth:`create`, :meth:`exercise`,
        :meth:`exercise_by_key`, or :meth:`create_and_exercise`, as they are available over both
        the gRPC Ledger API and HTTP JSON API; additionally those methods can provide more
        information about what happened.

        This method can be used to submit multiple disparate commands as a single transaction, but
        if you find yourself needing to do this, you may want to consider moving more of your logic
        into Daml so that only a single command is needed from the outside in order to satisfy your
        use case.
        """
        if __commands is None:
            return
        elif isinstance(__commands, Command):
            __commands = [__commands]

        stub = lapipb.CommandServiceStub(self.channel)

        meta = self._command_meta(
            workflow_id=workflow_id, command_id=command_id, read_as=read_as, act_as=act_as
        )
        commands = await asyncio.gather(*map(self._codec.encode_command, __commands))
        request = self._submit_and_wait_request(commands, meta)
        await stub.SubmitAndWait(request)

    async def create(
        self,
        __template_id: Union[str, TypeConName],
        __payload: ContractData,
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
        read_as: Union[None, Party, Collection[Party]] = None,
        act_as: Union[None, Party, Collection[Party]] = None,
    ) -> CreateEvent:
        """
        Create a contract for a given template.

        :param __template_id:
            The template of the contract to be created.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __payload:
            Template arguments for the contract to be created.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param workflow_id:
            An optional workflow ID.
        :param command_id:
            An optional command ID. If unspecified, a random one will be created.
        :param read_as:
            An optional set of read-as parties to use to submit this command. Note that for a
            ledger with authorization, these parties must be a subset of the parties in the token.
        :param act_as:
            An optional set of act-as parties to use to submit this command. Note that for a
            ledger with authorization, these parties must be a subset of the parties in the token.
        :return:
            The :class:`CreateEvent` that represents the contract that was successfully created.
        """
        stub = lapipb.CommandServiceStub(self.channel)

        meta = self._command_meta(
            workflow_id=workflow_id, command_id=command_id, read_as=read_as, act_as=act_as
        )
        commands = [
            lapipb.Command(create=await self._codec.encode_create_command(__template_id, __payload))
        ]
        request = self._submit_and_wait_request(commands, meta)
        response = await stub.SubmitAndWaitForTransaction(request)

        return await self._codec.decode_created_event(response.transaction.events[0].created)

    async def exercise(
        self,
        __contract_id: ContractId,
        __choice_name: str,
        __argument: Optional[ContractData] = None,
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
        read_as: Union[None, Party, Collection[Party]] = None,
        act_as: Union[None, Party, Collection[Party]] = None,
    ) -> ExerciseResponse:
        """
        Exercise a choice on a contract identified by its contract ID.

        :param __contract_id:
            The contract ID of the contract to exercise.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __choice_name:
            The name of the choice to exercise.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __argument:
            The choice arguments. Can be omitted for choices that take no argument.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param workflow_id:
            An optional workflow ID.
        :param command_id:
            An optional command ID. If unspecified, a random one will be created.
        :param read_as:
            An optional set of read-as parties to use to submit this command. Note that for a
            ledger with authorization, these parties must be a subset of the parties in the token.
        :param act_as:
            An optional set of act-as parties to use to submit this command. Note that for a
            ledger with authorization, these parties must be a subset of the parties in the token.
        :return:
            The return value of the choice, together with a list of events that occurred as a result
            of exercising the choice.
        """
        stub = lapipb.CommandServiceStub(self.channel)

        meta = self._command_meta(
            workflow_id=workflow_id, command_id=command_id, read_as=read_as, act_as=act_as
        )
        commands = [
            lapipb.Command(
                exercise=await self._codec.encode_exercise_command(
                    __contract_id, __choice_name, __argument
                )
            )
        ]
        request = self._submit_and_wait_request(commands, meta)
        response = await stub.SubmitAndWaitForTransactionTree(request)

        return await self._codec.decode_exercise_response(response.transaction)

    async def create_and_exercise(
        self,
        __template_id: Union[str, TypeConName],
        __payload: ContractData,
        __choice_name: str,
        __argument: Optional[ContractData] = None,
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
        read_as: Union[None, Party, Collection[Party]] = None,
        act_as: Union[None, Party, Collection[Party]] = None,
    ) -> ExerciseResponse:
        """
        Exercise a choice on a newly-created contract, in a single transaction.

        :param __template_id:
            The template of the contract to be created (positional argument only).

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __payload:
            Template arguments for the contract to be created (positional argument only).

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __choice_name:
            The name of the choice to exercise (positional argument only).

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __argument:
            The choice arguments. Can be omitted for choices that take no argument (positional
            argument only).

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param workflow_id:
            An optional workflow ID.
        :param command_id:
            An optional command ID. If unspecified, a random one will be created.
        :param read_as:
            An optional set of read-as parties to use to submit this command. Note that for a
            ledger with authorization, these parties must be a subset of the parties in the token.
        :param act_as:
            An optional set of act-as parties to use to submit this command. Note that for a
            ledger with authorization, these parties must be a subset of the parties in the token.
        :return:
            The return value of the choice, together with a list of events that occurred as a result
            of exercising the choice.
        """
        stub = lapipb.CommandServiceStub(self.channel)

        meta = self._command_meta(
            workflow_id=workflow_id, command_id=command_id, read_as=read_as, act_as=act_as
        )
        commands = [
            lapipb.Command(
                createAndExercise=await self._codec.encode_create_and_exercise_command(
                    __template_id, __payload, __choice_name, __argument
                )
            )
        ]
        request = self._submit_and_wait_request(commands, meta)
        response = await stub.SubmitAndWaitForTransactionTree(request)

        return await self._codec.decode_exercise_response(response.transaction)

    async def exercise_by_key(
        self,
        __template_id: Union[str, TypeConName],
        __choice_name: str,
        __key: Any,
        __argument: Optional[ContractData] = None,
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
        read_as: Union[None, Party, Collection[Party]] = None,
        act_as: Union[None, Party, Collection[Party]] = None,
    ) -> "ExerciseResponse":
        """
        Exercise a choice on a contract identified by its contract key.

        :param __template_id:
            The template of the contract to be created (positional argument only).

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __choice_name:
            The name of the choice to exercise.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __key:
            The key of the contract to exercise.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __argument:
            The choice arguments. Can be omitted for choices that take no argument.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param workflow_id:
            An optional workflow ID.
        :param command_id:
            An optional command ID. If unspecified, a random one will be created.
        :param read_as:
            An optional set of read-as parties to use to submit this command. Note that for a
            ledger with authorization, these parties must be a subset of the parties in the token.
        :param act_as:
            An optional set of act-as parties to use to submit this command. Note that for a
            ledger with authorization, these parties must be a subset of the parties in the token.
        :return:
            The return value of the choice, together with a list of events that occurred as a result
            of exercising the choice.
        """
        stub = lapipb.CommandServiceStub(self.channel)

        meta = self._command_meta(
            workflow_id=workflow_id, command_id=command_id, read_as=read_as, act_as=act_as
        )
        commands = [
            lapipb.Command(
                exerciseByKey=await self._codec.encode_exercise_by_key_command(
                    __template_id, __choice_name, __key, __argument
                )
            )
        ]
        request = self._submit_and_wait_request(commands, meta)
        response = await stub.SubmitAndWaitForTransactionTree(request)

        return await self._codec.decode_exercise_response(response.transaction)

    async def archive(
        self,
        __contract_id: ContractId,
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
        read_as: Union[None, Party, Collection[Party]] = None,
        act_as: Union[None, Party, Collection[Party]] = None,
    ) -> ArchiveEvent:
        """
        Archive a choice on a contract identified by its contract ID.

        :param __contract_id:
            The contract ID of the contract to exercise.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param workflow_id:
            An optional workflow ID.
        :param command_id:
            An optional command ID. If unspecified, a random one will be created.
        :param read_as:
            An optional set of read-as parties to use to submit this command. Note that for a
            ledger with authorization, these parties must be a subset of the parties in the token.
        :param act_as:
            An optional set of act-as parties to use to submit this command. Note that for a
            ledger with authorization, these parties must be a subset of the parties in the token.
        :return:
            The return value of the choice, together with a list of events that occurred as a result
            of exercising the choice.
        """
        await self.exercise(
            __contract_id,
            "Archive",
            workflow_id=workflow_id,
            command_id=command_id,
            read_as=read_as,
            act_as=act_as,
        )
        return ArchiveEvent(__contract_id)

    async def archive_by_key(
        self,
        __template_id: str,
        __key: Any,
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
        read_as: Union[None, Party, Collection[Party]] = None,
        act_as: Union[None, Party, Collection[Party]] = None,
    ) -> ArchiveEvent:
        """
        Exercise a choice on a contract identified by its contract key.

        :param __template_id:
            The template of the contract to be created (positional argument only).

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __key:
            The key of the contract to exercise.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param workflow_id:
            An optional workflow ID.
        :param command_id:
            An optional command ID. If unspecified, a random one will be created.
        :param read_as:
            An optional set of read-as parties to use to submit this command. Note that for a
            ledger with authorization, these parties must be a subset of the parties in the token.
        :param act_as:
            An optional set of act-as parties to use to submit this command. Note that for a
            ledger with authorization, these parties must be a subset of the parties in the token.
        :return:
            The return value of the choice, together with a list of events that occurred as a result
            of exercising the choice.
        """
        response = await self.exercise_by_key(
            __template_id,
            "Archive",
            __key,
            workflow_id=workflow_id,
            command_id=command_id,
            read_as=read_as,
            act_as=act_as,
        )
        return next(iter(event for event in response.events if isinstance(event, ArchiveEvent)))

    def _submit_and_wait_request(
        self,
        commands: Collection[lapipb.Command],
        meta: CommandMeta,
    ) -> lapipb.SubmitAndWaitRequest:
        # this is support for versions of Daml Connect prior to 1.9.0
        act_as = meta.act_as
        if act_as is not None:
            act_as_party = act_as[0] if act_as else None
        else:
            raise ValueError("current access rights do not include any act-as parties")

        return lapipb.SubmitAndWaitRequest(
            commands=lapipb.Commands(
                ledger_id=self._config.access.ledger_id,
                application_id=self._config.access.application_name,
                command_id=meta.command_id,
                workflow_id=meta.workflow_id,
                party=act_as_party,
                commands=commands,
                act_as=meta.act_as,
                read_as=meta.read_as,
            )
        )

    def _command_meta(
        self,
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
        read_as: Union[None, Party, Collection[Party]] = None,
        act_as: Union[None, Party, Collection[Party]] = None,
    ):
        read_as = self._read_as(read_as)
        if act_as is None:
            act_as = self._config.access.act_as

        return CommandMeta(
            workflow_id=workflow_id, command_id=command_id, read_as=read_as, act_as=act_as
        )

    # endregion

    # region Read API

    async def get_ledger_end(self) -> "str":
        """
        Return the offset at the end of the ledger.
        """
        stub = lapipb.TransactionServiceStub(self.channel)

        request = lapipb.GetLedgerEndRequest(ledger_id=self._config.access.ledger_id)
        response = await stub.GetLedgerEnd(request)
        return response.offset.absolute

    def query(
        self,
        __template_id: Union[str, TypeConName] = "*",
        __query: Query = None,
        *,
        begin_offset: Optional[str] = None,
        end_offset: Optional[str] = None,
        read_as: Union[None, Party, Collection[Party]] = None,
    ) -> "QueryStream":
        """
        Return the create events from the active contract set service as a stream.

        If you find yourself repeatedly calling :meth:`query` or :meth:`query_many` over the same
        set of templates, you may want to consider :class:`ACS` instead, which is a utility class
        that helps you maintain a "live" state of the ACS.

        :param __template_id:
            The name of the template for which to fetch contracts.
        :param __query:
            A filter to apply to the set of returned contracts.
        :param begin_offset:
            The starting offset at which to read an active contract set. If ``None``, contracts
            are read from the beginning, and using the Active Contract Set Service instead of the
            Transaction Service.
        :param end_offset:
            The ending offset. If ``None``, contracts are read until the end of the stream.
            In order to read indefinitely, use :meth:`stream` instead.
        :param read_as:
            An optional set of read-as parties to use to submit this query. Note that for a
            ledger with authorization, these parties must be a subset of the parties in the token.
        """
        offset = LedgerOffsetRange(begin_offset, end_offset if end_offset is not None else END)
        return QueryStream(
            self,
            parse_query({__template_id: __query}, server_side_filters=False),
            offset,
            self._read_as(read_as),
        )

    def query_many(
        self,
        *queries: Queries,
        begin_offset: Optional[str] = None,
        end_offset: Optional[str] = None,
        read_as: Union[None, Party, Collection[Party]] = None,
    ) -> "QueryStream":
        """
        Return the create events from the active contract set service as a stream.

        If you find yourself repeatedly calling :meth:`query` or :meth:`query_many` over the same
        set of templates, you may want to consider :class:`ACS` instead, which is a utility class
        that helps you maintain a "live" state of the ACS.

        :param queries:
            A map of template IDs to filter to apply to the set of returned contracts.
        :param begin_offset:
            The starting offset at which to read an active contract set. If ``None``, contracts
            are read from the beginning, and using the Active Contract Set Service instead of the
            Transaction Service.
        :param end_offset:
            The ending offset. If ``None``, contracts are read until the end of the stream.
            In order to read indefinitely, use :meth:`stream_many` instead.
        :param read_as:
            An optional set of read-as parties to use to submit this query. Note that for a
            ledger with authorization, these parties must be a subset of the parties in the token.
        """
        offset = LedgerOffsetRange(begin_offset, end_offset if end_offset is not None else END)
        return QueryStream(
            self, parse_query(*queries, server_side_filters=False), offset, self._read_as(read_as)
        )

    def stream(
        self,
        __template_id: Union[str, TypeConName] = "*",
        __query: Query = None,
        *,
        offset: Optional[str] = None,
        read_as: Union[None, Party, Collection[Party]] = None,
    ) -> "QueryStream":
        """
        Stream create/archive events.

        When ``offset`` is ``None``, create events from the active contract set are returned first,
        followed by a continuous stream of updates (creates/archives).

        Otherwise, ``offset`` can be supplied to resume a stream from a prior point where a
        ``Boundary`` was returned from a previous stream.

        :param __template_id:
            The name of the template for which to fetch contracts.
        :param __query:
            A filter to apply to the set of returned contracts. Note that this does not filter
            :class:`ArchiveEvent`; readers of the stream MUST be able to cope with "mismatched"
            archives that come from the result of applying a filter.
        :param offset:
            An optional offset at which to start receiving events. If ``None``, start from the
            beginning.
        :param read_as:
            An optional set of read-as parties to use to submit this query. Note that for a
            ledger with authorization, these parties must be a subset of the parties in the token.
        """
        return QueryStream(
            self,
            parse_query({__template_id: __query}, server_side_filters=False),
            from_offset_until_forever(offset),
            self._read_as(read_as),
        )

    def stream_many(
        self,
        *queries: Queries,
        offset: Optional[str] = None,
        read_as: Union[None, Party, Collection[Party]] = None,
    ) -> "QueryStream":
        """
        Stream create/archive events from more than one template ID in the same stream.

        When ``offset`` is ``None``, create events from the active contract set are returned first,
        followed by a continuous stream of updates (creates/archives).

        Otherwise, ``offset`` can be supplied to resume a stream from a prior point where a
        ``Boundary`` was returned from a previous stream.

        :param queries:
            A map of template IDs to filter to apply to the set of returned contracts. Note that
            this does not filter :class:`ArchiveEvent`; readers of the stream MUST be able to cope
            with "mismatched" archives that come from the result of applying a filter.
        :param offset:
            An optional offset at which to start receiving events. If ``None``, start from the
            beginning.
        :param read_as:
            An optional set of read-as parties to use to submit this query. Note that for a
            ledger with authorization, these parties must be a subset of the parties in the token.
        """
        return QueryStream(
            self,
            parse_query(*queries, server_side_filters=False),
            from_offset_until_forever(offset),
            self._read_as(read_as),
        )

    def _read_as(self, read_as: Union[None, Party, Collection[Party]] = None) -> Collection[Party]:
        if read_as is None:
            return self.config.access.read_as
        else:
            return to_parties(read_as)

    # endregion

    async def get_version(self) -> "Version":
        stub = lapipb.VersionServiceStub(self.channel)
        request = lapipb.GetLedgerApiVersionRequest(ledger_id=self._config.access.ledger_id)
        response = await stub.GetLedgerApiVersion(request)
        return Codec.decode_version(response)

    # region User Management calls

    async def get_user(self, user_id: "Optional[str]" = None) -> "User":
        stub = lapiadminpb.UserManagementServiceStub(self.channel)
        request = lapiadminpb.GetUserRequest(user_id=user_id)
        response = await stub.GetUser(request)
        return Codec.decode_user(response.user)

    async def create_user(self, user: "User", rights: "Optional[Sequence[Right]]" = None) -> "User":
        stub = lapiadminpb.UserManagementServiceStub(self.channel)
        request = lapiadminpb.CreateUserRequest(user=Codec.encode_user(user))
        if rights is not None:
            request.rights.extend(Codec.encode_right(right) for right in rights)

        response = await stub.CreateUser(request)
        return Codec.decode_user(response.user)

    async def list_users(self) -> "Sequence[User]":
        stub = lapiadminpb.UserManagementServiceStub(self.channel)
        request = lapiadminpb.ListUsersRequest()
        response = await stub.ListUsers(request)
        return [Codec.decode_user(user) for user in response.users]

    async def list_user_rights(self, user_id: "Optional[str]" = None) -> "Sequence[Right]":
        stub = lapiadminpb.UserManagementServiceStub(self.channel)
        request = lapiadminpb.ListUserRightsRequest(user_id=user_id)
        response = await stub.ListUserRights(request)
        return [Codec.decode_right(user) for user in response.rights]

    # endregion

    # region Party Management calls

    async def allocate_party(
        self, *, identifier_hint: str = None, display_name: str = None
    ) -> "PartyInfo":
        """
        Allocate a new party.
        """
        stub = lapiadminpb.PartyManagementServiceStub(self.channel)
        request = lapiadminpb.AllocatePartyRequest(
            party_id_hint=Party(identifier_hint) if identifier_hint else None,
            display_name=display_name,
        )

        response = await retry(lambda: stub.AllocateParty(request))
        return Codec.decode_party_info(response.party_details)

    async def list_known_parties(self) -> Sequence[PartyInfo]:
        stub = lapiadminpb.PartyManagementServiceStub(self.channel)
        request = lapiadminpb.ListKnownPartiesRequest()
        response = await stub.ListKnownParties(request)
        return [Codec.decode_party_info(pd) for pd in response.party_details]

    # endregion

    # region Package Management calls

    async def get_package(self, package_id: PackageRef) -> bytes:
        stub = lapipb.PackageServiceStub(self.channel)
        request = lapipb.GetPackageRequest(
            ledger_id=self._config.access.ledger_id, package_id=package_id
        )
        response = await stub.GetPackage(request)
        return response.archive_payload

    async def list_package_ids(self) -> AbstractSet[PackageRef]:
        stub = lapipb.PackageServiceStub(self.channel)
        request = lapipb.ListPackagesRequest(ledger_id=self._config.access.ledger_id)
        response = await stub.ListPackages(request)
        return frozenset({PackageRef(pkg_id) for pkg_id in response.package_ids})

    async def upload_package(self, contents: bytes) -> None:
        stub = lapiadminpb.PackageManagementServiceStub(self.channel)
        request = lapiadminpb.UploadDarFileRequest(dar_file=contents)
        await stub.UploadDarFile(request)
        return

    # endregion

    # region Metering Report calls

    async def get_metering_report(
        self, from_: datetime, to: Optional[datetime] = None, application_id: Optional[str] = None
    ) -> ParticipantMeteringReport:
        stub = lapiadminpb.MeteringReportServiceStub(self.channel)

        # This slightly awkward construction is due to Protobuf's Python type not escaping field
        # names that happen to match keywords. Unfortunately mypy can't follow what is going on
        # when we do this, so we must be careful!
        kwargs = {"from": datetime_to_timestamp(from_)}  # type: Dict[str, Any]
        if to is not None:
            kwargs["to"] = datetime_to_timestamp(to)
        if application_id is not None:
            kwargs["application_id"] = application_id
        request = lapiadminpb.GetMeteringReportRequest(**kwargs)  # type: ignore
        response = await stub.GetMeteringReport(request)
        return Codec.decode_get_metering_report_response(response)

    # end region


class QueryStream(aio.QueryStreamBase):
    def __init__(
        self,
        conn: Connection,
        filters: Optional[Mapping[TypeConName, Filter]],
        offset_range: LedgerOffsetRange,
        read_as: Collection[Party],
    ):
        self.conn = conn
        self._filters = filters
        self._offset_range = offset_range
        self._read_as = read_as
        self._response_stream = None  # type: Optional[UnaryStreamCall]
        self._closed = False

    @property
    def is_closed(self) -> bool:
        return self._closed

    async def close(self) -> None:
        # make sure to mark the object as "closed"; when we're closed, we don't mind cancellation
        # errors, because we're the one triggering the cancellation
        self._closed = True
        if self._response_stream is not None:
            self._response_stream.cancel()
            self._response_stream = None

    async def items(self):
        """
        Return an asynchronous stream of events.

        .. code-block:: python

            async with conn.query('SampleApp:Iou') as query:
                async for r in query:
                    print(f"Offset: {r.offset}")
                    for event in r.events:
                        print(f"  Event: {event}")

        :return:
            A stream of responses, where each response contains one or more events at a particular
            offset.

            At least one initial :class:`Boundary` is always returned, even if the stream is empty.
            In this case, the first returned object is a :class:`Boundary` with ``offset=None``.
        """
        log = self.conn.config.logger
        async with _translate_exceptions(self.conn), self, _allow_cancel(lambda: self._closed):
            filters = await self.conn.codec.encode_filters(self._filters)
            filters_by_party = {party: filters for party in self._read_as}
            tx_filter_pb = lapipb.TransactionFilter(filters_by_party=filters_by_party)

            offset = self._offset_range.begin
            if offset:
                log.debug("Skipped reading from the ACS because begin offset is %r", offset)
            else:
                # when starting from the beginning of the ledger, the Active Contract Set service
                # lets us catch up more quickly than having to parse every create/archive event
                # ourselves
                log.debug("Reading from the ACS...")
                async for event in self._acs_events(tx_filter_pb):
                    if isinstance(event, CreateEvent):
                        await self._emit_create(event)
                    elif isinstance(event, Boundary):
                        offset = event.offset
                        await self._emit_boundary(event)
                    else:
                        warnings.warn(f"Received an unknown event: {event}", ProtocolWarning)
                    yield event

                # when reading from the Active Contract Set service, if we're supposed to stop
                # at "the end", then the Active Contract Set data is all that we'll return
                if self._offset_range.end == END:
                    log.debug(
                        "Not reading from transaction stream because we were only asked for a snapshot."
                    )
                    return

            # now start returning events as they come off the transaction stream; note this
            # stream will never naturally close, so it's on the caller to call close() or to
            # otherwise exit our current context
            log.debug("Reading a transaction stream: %s", self._offset_range)
            async for event in self._tx_events(tx_filter_pb, offset, self._offset_range.end):
                log.debug("Received an event: %s", event)
                if isinstance(event, CreateEvent):
                    await self._emit_create(event)
                elif isinstance(event, ArchiveEvent):
                    await self._emit_archive(event)
                elif isinstance(event, Boundary):
                    await self._emit_boundary(event)
                else:
                    warnings.warn(f"Received an unknown event: {event}", ProtocolWarning)
                yield event

    async def _acs_events(
        self, filter_pb: lapipb.TransactionFilter
    ) -> AsyncIterable[Union[CreateEvent, Boundary]]:
        stub = lapipb.ActiveContractsServiceStub(self.conn.channel)

        request = lapipb.GetActiveContractsRequest(
            ledger_id=self.conn.config.access.ledger_id, filter=filter_pb
        )
        self._response_stream = response_stream = stub.GetActiveContracts(request)

        offset = None
        async for response in response_stream:
            LOG.debug(
                "ACS start (offset %r, %d event(s))",
                response.offset,
                len(response.active_contracts),
            )
            for event in response.active_contracts:
                c_evt = await self.conn.codec.decode_created_event(event)
                if self._is_match(c_evt):
                    yield c_evt
            # for ActiveContractSetResponse messages, only the last offset is actually relevant
            LOG.debug("ACS end (offset %r)", response.offset)
            offset = response.offset
        yield Boundary(offset)

    async def _tx_events(
        self,
        filter_pb: lapipb.TransactionFilter,
        begin_offset: Optional[str],
        end_offset: "Union[None, str, End]",
    ) -> AsyncIterable[Union[CreateEvent, ArchiveEvent, Boundary]]:
        stub = lapipb.TransactionServiceStub(self.conn.channel)

        request = lapipb.GetTransactionsRequest(
            ledger_id=self.conn.config.access.ledger_id,
            filter=filter_pb,
            begin=self.conn.codec.encode_begin_offset(begin_offset),
            end=self.conn.codec.encode_end_offset(end_offset),
        )

        self._response_stream = response_stream = stub.GetTransactions(request)
        async for response in response_stream:
            for tx in response.transactions:
                for event in tx.events:
                    event_type = event.WhichOneof("event")
                    if event_type == "created":
                        c_evt = await self.conn.codec.decode_created_event(event.created)
                        if self._is_match(c_evt):
                            yield c_evt
                    elif event_type == "archived":
                        yield await self.conn.codec.decode_archived_event(event.archived)
                    else:
                        warnings.warn(f"Unknown Event({event_type}=...)", ProtocolWarning)
                yield Boundary(tx.offset)

    def _is_match(self, event: CreateEvent) -> bool:
        # if there are no filters, then everything is a match
        if self._filters is None:
            return True

        for name, f in self._filters.items():
            # for each filter, if this contract type could be interpreted as a match for the current
            # filter, then apply a client-side filter
            if is_match(name, event.contract_id.value_type):
                return f.client_side is None or f.client_side(event.payload)

        # we checked every name, but none of them matched
        return False



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/testing/reset_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/daml/ledger/api/v1/testing/reset_service.proto',
  package='com.daml.ledger.api.v1.testing',
  syntax='proto3',
  serialized_options=b'\n\036com.daml.ledger.api.v1.testingB\026ResetServiceOuterClassZWgithub.com/digital-asset/dazl-client/go/v7/pkg/generated/com/daml/ledger/api/v1/testing\252\002\036Com.Daml.Ledger.Api.V1.Testing',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n2com/daml/ledger/api/v1/testing/reset_service.proto\x12\x1e\x63om.daml.ledger.api.v1.testing\x1a\x1bgoogle/protobuf/empty.proto\"+\n\x0cResetRequest\x12\x1b\n\tledger_id\x18\x01 \x01(\tR\x08ledgerId2]\n\x0cResetService\x12M\n\x05Reset\x12,.com.daml.ledger.api.v1.testing.ResetRequest\x1a\x16.google.protobuf.EmptyB\xb2\x01\n\x1e\x63om.daml.ledger.api.v1.testingB\x16ResetServiceOuterClassZWgithub.com/digital-asset/dazl-client/go/v7/pkg/generated/com/daml/ledger/api/v1/testing\xaa\x02\x1e\x43om.Daml.Ledger.Api.V1.Testingb\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,])




_RESETREQUEST = _descriptor.Descriptor(
  name='ResetRequest',
  full_name='com.daml.ledger.api.v1.testing.ResetRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ledger_id', full_name='com.daml.ledger.api.v1.testing.ResetRequest.ledger_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='ledgerId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=115,
  serialized_end=158,
)

DESCRIPTOR.message_types_by_name['ResetRequest'] = _RESETREQUEST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ResetRequest = _reflection.GeneratedProtocolMessageType('ResetRequest', (_message.Message,), {
  'DESCRIPTOR' : _RESETREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.testing.reset_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.testing.ResetRequest)
  })
_sym_db.RegisterMessage(ResetRequest)


DESCRIPTOR._options = None

_RESETSERVICE = _descriptor.ServiceDescriptor(
  name='ResetService',
  full_name='com.daml.ledger.api.v1.testing.ResetService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=160,
  serialized_end=253,
  methods=[
  _descriptor.MethodDescriptor(
    name='Reset',
    full_name='com.daml.ledger.api.v1.testing.ResetService.Reset',
    index=0,
    containing_service=None,
    input_type=_RESETREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_RESETSERVICE)

DESCRIPTOR.services_by_name['ResetService'] = _RESETSERVICE

# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/experimental_features.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2com/daml/ledger/api/v1/experimental_features.proto\x12\x16\x63om.daml.ledger.api.v1\"\xe9\x04\n\x14\x45xperimentalFeatures\x12v\n\x18self_service_error_codes\x18\x01 \x01(\x0b\x32\x39.com.daml.ledger.api.v1.ExperimentalSelfServiceErrorCodesB\x02\x18\x01R\x15selfServiceErrorCodes\x12O\n\x0bstatic_time\x18\x02 \x01(\x0b\x32..com.daml.ledger.api.v1.ExperimentalStaticTimeR\nstaticTime\x12i\n\x15\x63ommand_deduplication\x18\x03 \x01(\x0b\x32\x34.com.daml.ledger.api.v1.CommandDeduplicationFeaturesR\x14\x63ommandDeduplication\x12\x62\n\x12optional_ledger_id\x18\x04 \x01(\x0b\x32\x34.com.daml.ledger.api.v1.ExperimentalOptionalLedgerIdR\x10optionalLedgerId\x12R\n\x0c\x63ontract_ids\x18\x05 \x01(\x0b\x32/.com.daml.ledger.api.v1.ExperimentalContractIdsR\x0b\x63ontractIds\x12\x65\n\x13\x63ommitter_event_log\x18\x06 \x01(\x0b\x32\x35.com.daml.ledger.api.v1.ExperimentalCommitterEventLogR\x11\x63ommitterEventLog\"\'\n!ExperimentalSelfServiceErrorCodes:\x02\x18\x01\"6\n\x16\x45xperimentalStaticTime\x12\x1c\n\tsupported\x18\x01 \x01(\x08R\tsupported\"\xcb\x02\n\x1c\x43ommandDeduplicationFeatures\x12{\n\x1c\x64\x65\x64uplication_period_support\x18\x01 \x01(\x0b\x32\x39.com.daml.ledger.api.v1.CommandDeduplicationPeriodSupportR\x1a\x64\x65\x64uplicationPeriodSupport\x12_\n\x12\x64\x65\x64uplication_type\x18\x02 \x01(\x0e\x32\x30.com.daml.ledger.api.v1.CommandDeduplicationTypeR\x11\x64\x65\x64uplicationType\x12M\n#max_deduplication_duration_enforced\x18\x03 \x01(\x08R maxDeduplicationDurationEnforced\"\x1e\n\x1c\x45xperimentalOptionalLedgerId\"\xbf\x03\n!CommandDeduplicationPeriodSupport\x12n\n\x0eoffset_support\x18\x01 \x01(\x0e\x32G.com.daml.ledger.api.v1.CommandDeduplicationPeriodSupport.OffsetSupportR\roffsetSupport\x12t\n\x10\x64uration_support\x18\x02 \x01(\x0e\x32I.com.daml.ledger.api.v1.CommandDeduplicationPeriodSupport.DurationSupportR\x0f\x64urationSupport\"d\n\rOffsetSupport\x12\x18\n\x14OFFSET_NOT_SUPPORTED\x10\x00\x12\x19\n\x15OFFSET_NATIVE_SUPPORT\x10\x01\x12\x1e\n\x1aOFFSET_CONVERT_TO_DURATION\x10\x02\"N\n\x0f\x44urationSupport\x12\x1b\n\x17\x44URATION_NATIVE_SUPPORT\x10\x00\x12\x1e\n\x1a\x44URATION_CONVERT_TO_OFFSET\x10\x01\"\xa5\x01\n\x17\x45xperimentalContractIds\x12S\n\x02v1\x18\x01 \x01(\x0e\x32\x43.com.daml.ledger.api.v1.ExperimentalContractIds.ContractIdV1SupportR\x02v1\"5\n\x13\x43ontractIdV1Support\x12\x0c\n\x08SUFFIXED\x10\x00\x12\x10\n\x0cNON_SUFFIXED\x10\x01\"\xcd\x01\n\x1d\x45xperimentalCommitterEventLog\x12q\n\x0e\x65vent_log_type\x18\x01 \x01(\x0e\x32K.com.daml.ledger.api.v1.ExperimentalCommitterEventLog.CommitterEventLogTypeR\x0c\x65ventLogType\"9\n\x15\x43ommitterEventLogType\x12\x0f\n\x0b\x43\x45NTRALIZED\x10\x00\x12\x0f\n\x0b\x44ISTRIBUTED\x10\x01*I\n\x18\x43ommandDeduplicationType\x12\x0e\n\nASYNC_ONLY\x10\x00\x12\x1d\n\x19\x41SYNC_AND_CONCURRENT_SYNC\x10\x01\x42\x98\x01\n\x16\x63om.daml.ledger.api.v1B\x1e\x45xperimentalFeaturesOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\xaa\x02\x16\x43om.Daml.Ledger.Api.V1b\x06proto3')

_COMMANDDEDUPLICATIONTYPE = DESCRIPTOR.enum_types_by_name['CommandDeduplicationType']
CommandDeduplicationType = enum_type_wrapper.EnumTypeWrapper(_COMMANDDEDUPLICATIONTYPE)
ASYNC_ONLY = 0
ASYNC_AND_CONCURRENT_SYNC = 1


_EXPERIMENTALFEATURES = DESCRIPTOR.message_types_by_name['ExperimentalFeatures']
_EXPERIMENTALSELFSERVICEERRORCODES = DESCRIPTOR.message_types_by_name['ExperimentalSelfServiceErrorCodes']
_EXPERIMENTALSTATICTIME = DESCRIPTOR.message_types_by_name['ExperimentalStaticTime']
_COMMANDDEDUPLICATIONFEATURES = DESCRIPTOR.message_types_by_name['CommandDeduplicationFeatures']
_EXPERIMENTALOPTIONALLEDGERID = DESCRIPTOR.message_types_by_name['ExperimentalOptionalLedgerId']
_COMMANDDEDUPLICATIONPERIODSUPPORT = DESCRIPTOR.message_types_by_name['CommandDeduplicationPeriodSupport']
_EXPERIMENTALCONTRACTIDS = DESCRIPTOR.message_types_by_name['ExperimentalContractIds']
_EXPERIMENTALCOMMITTEREVENTLOG = DESCRIPTOR.message_types_by_name['ExperimentalCommitterEventLog']
_COMMANDDEDUPLICATIONPERIODSUPPORT_OFFSETSUPPORT = _COMMANDDEDUPLICATIONPERIODSUPPORT.enum_types_by_name['OffsetSupport']
_COMMANDDEDUPLICATIONPERIODSUPPORT_DURATIONSUPPORT = _COMMANDDEDUPLICATIONPERIODSUPPORT.enum_types_by_name['DurationSupport']
_EXPERIMENTALCONTRACTIDS_CONTRACTIDV1SUPPORT = _EXPERIMENTALCONTRACTIDS.enum_types_by_name['ContractIdV1Support']
_EXPERIMENTALCOMMITTEREVENTLOG_COMMITTEREVENTLOGTYPE = _EXPERIMENTALCOMMITTEREVENTLOG.enum_types_by_name['CommitterEventLogType']
ExperimentalFeatures = _reflection.GeneratedProtocolMessageType('ExperimentalFeatures', (_message.Message,), {
  'DESCRIPTOR' : _EXPERIMENTALFEATURES,
  '__module__' : 'com.daml.ledger.api.v1.experimental_features_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.ExperimentalFeatures)
  })
_sym_db.RegisterMessage(ExperimentalFeatures)

ExperimentalSelfServiceErrorCodes = _reflection.GeneratedProtocolMessageType('ExperimentalSelfServiceErrorCodes', (_message.Message,), {
  'DESCRIPTOR' : _EXPERIMENTALSELFSERVICEERRORCODES,
  '__module__' : 'com.daml.ledger.api.v1.experimental_features_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.ExperimentalSelfServiceErrorCodes)
  })
_sym_db.RegisterMessage(ExperimentalSelfServiceErrorCodes)

ExperimentalStaticTime = _reflection.GeneratedProtocolMessageType('ExperimentalStaticTime', (_message.Message,), {
  'DESCRIPTOR' : _EXPERIMENTALSTATICTIME,
  '__module__' : 'com.daml.ledger.api.v1.experimental_features_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.ExperimentalStaticTime)
  })
_sym_db.RegisterMessage(ExperimentalStaticTime)

CommandDeduplicationFeatures = _reflection.GeneratedProtocolMessageType('CommandDeduplicationFeatures', (_message.Message,), {
  'DESCRIPTOR' : _COMMANDDEDUPLICATIONFEATURES,
  '__module__' : 'com.daml.ledger.api.v1.experimental_features_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.CommandDeduplicationFeatures)
  })
_sym_db.RegisterMessage(CommandDeduplicationFeatures)

ExperimentalOptionalLedgerId = _reflection.GeneratedProtocolMessageType('ExperimentalOptionalLedgerId', (_message.Message,), {
  'DESCRIPTOR' : _EXPERIMENTALOPTIONALLEDGERID,
  '__module__' : 'com.daml.ledger.api.v1.experimental_features_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.ExperimentalOptionalLedgerId)
  })
_sym_db.RegisterMessage(ExperimentalOptionalLedgerId)

CommandDeduplicationPeriodSupport = _reflection.GeneratedProtocolMessageType('CommandDeduplicationPeriodSupport', (_message.Message,), {
  'DESCRIPTOR' : _COMMANDDEDUPLICATIONPERIODSUPPORT,
  '__module__' : 'com.daml.ledger.api.v1.experimental_features_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.CommandDeduplicationPeriodSupport)
  })
_sym_db.RegisterMessage(CommandDeduplicationPeriodSupport)

ExperimentalContractIds = _reflection.GeneratedProtocolMessageType('ExperimentalContractIds', (_message.Message,), {
  'DESCRIPTOR' : _EXPERIMENTALCONTRACTIDS,
  '__module__' : 'com.daml.ledger.api.v1.experimental_features_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.ExperimentalContractIds)
  })
_sym_db.RegisterMessage(ExperimentalContractIds)

ExperimentalCommitterEventLog = _reflection.GeneratedProtocolMessageType('ExperimentalCommitterEventLog', (_message.Message,), {
  'DESCRIPTOR' : _EXPERIMENTALCOMMITTEREVENTLOG,
  '__module__' : 'com.daml.ledger.api.v1.experimental_features_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.ExperimentalCommitterEventLog)
  })
_sym_db.RegisterMessage(ExperimentalCommitterEventLog)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026com.daml.ledger.api.v1B\036ExperimentalFeaturesOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\252\002\026Com.Daml.Ledger.Api.V1'
  _EXPERIMENTALFEATURES.fields_by_name['self_service_error_codes']._options = None
  _EXPERIMENTALFEATURES.fields_by_name['self_service_error_codes']._serialized_options = b'\030\001'
  _EXPERIMENTALSELFSERVICEERRORCODES._options = None
  _EXPERIMENTALSELFSERVICEERRORCODES._serialized_options = b'\030\001'
  _COMMANDDEDUPLICATIONTYPE._serialized_start=1987
  _COMMANDDEDUPLICATIONTYPE._serialized_end=2060
  _EXPERIMENTALFEATURES._serialized_start=79
  _EXPERIMENTALFEATURES._serialized_end=696
  _EXPERIMENTALSELFSERVICEERRORCODES._serialized_start=698
  _EXPERIMENTALSELFSERVICEERRORCODES._serialized_end=737
  _EXPERIMENTALSTATICTIME._serialized_start=739
  _EXPERIMENTALSTATICTIME._serialized_end=793
  _COMMANDDEDUPLICATIONFEATURES._serialized_start=796
  _COMMANDDEDUPLICATIONFEATURES._serialized_end=1127
  _EXPERIMENTALOPTIONALLEDGERID._serialized_start=1129
  _EXPERIMENTALOPTIONALLEDGERID._serialized_end=1159
  _COMMANDDEDUPLICATIONPERIODSUPPORT._serialized_start=1162
  _COMMANDDEDUPLICATIONPERIODSUPPORT._serialized_end=1609
  _COMMANDDEDUPLICATIONPERIODSUPPORT_OFFSETSUPPORT._serialized_start=1429
  _COMMANDDEDUPLICATIONPERIODSUPPORT_OFFSETSUPPORT._serialized_end=1529
  _COMMANDDEDUPLICATIONPERIODSUPPORT_DURATIONSUPPORT._serialized_start=1531
  _COMMANDDEDUPLICATIONPERIODSUPPORT_DURATIONSUPPORT._serialized_end=1609
  _EXPERIMENTALCONTRACTIDS._serialized_start=1612
  _EXPERIMENTALCONTRACTIDS._serialized_end=1777
  _EXPERIMENTALCONTRACTIDS_CONTRACTIDV1SUPPORT._serialized_start=1724
  _EXPERIMENTALCONTRACTIDS_CONTRACTIDV1SUPPORT._serialized_end=1777
  _EXPERIMENTALCOMMITTEREVENTLOG._serialized_start=1780
  _EXPERIMENTALCOMMITTEREVENTLOG._serialized_end=1985
  _EXPERIMENTALCOMMITTEREVENTLOG_COMMITTEREVENTLOGTYPE._serialized_start=1928
  _EXPERIMENTALCOMMITTEREVENTLOG_COMMITTEREVENTLOGTYPE._serialized_end=1985
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/testing/time_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n1com/daml/ledger/api/v1/testing/time_service.proto\x12\x1e\x63om.daml.ledger.api.v1.testing\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"-\n\x0eGetTimeRequest\x12\x1b\n\tledger_id\x18\x01 \x01(\tR\x08ledgerId\"P\n\x0fGetTimeResponse\x12=\n\x0c\x63urrent_time\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0b\x63urrentTime\"\xa3\x01\n\x0eSetTimeRequest\x12\x1b\n\tledger_id\x18\x01 \x01(\tR\x08ledgerId\x12=\n\x0c\x63urrent_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0b\x63urrentTime\x12\x35\n\x08new_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x07newTime2\xce\x01\n\x0bTimeService\x12l\n\x07GetTime\x12..com.daml.ledger.api.v1.testing.GetTimeRequest\x1a/.com.daml.ledger.api.v1.testing.GetTimeResponse0\x01\x12Q\n\x07SetTime\x12..com.daml.ledger.api.v1.testing.SetTimeRequest\x1a\x16.google.protobuf.EmptyB\xa7\x01\n\x1e\x63om.daml.ledger.api.v1.testingB\x15TimeServiceOuterClassZMgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1/testing\xaa\x02\x1e\x43om.Daml.Ledger.Api.V1.Testingb\x06proto3')



_GETTIMEREQUEST = DESCRIPTOR.message_types_by_name['GetTimeRequest']
_GETTIMERESPONSE = DESCRIPTOR.message_types_by_name['GetTimeResponse']
_SETTIMEREQUEST = DESCRIPTOR.message_types_by_name['SetTimeRequest']
GetTimeRequest = _reflection.GeneratedProtocolMessageType('GetTimeRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETTIMEREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.testing.time_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.testing.GetTimeRequest)
  })
_sym_db.RegisterMessage(GetTimeRequest)

GetTimeResponse = _reflection.GeneratedProtocolMessageType('GetTimeResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETTIMERESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.testing.time_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.testing.GetTimeResponse)
  })
_sym_db.RegisterMessage(GetTimeResponse)

SetTimeRequest = _reflection.GeneratedProtocolMessageType('SetTimeRequest', (_message.Message,), {
  'DESCRIPTOR' : _SETTIMEREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.testing.time_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.testing.SetTimeRequest)
  })
_sym_db.RegisterMessage(SetTimeRequest)

_TIMESERVICE = DESCRIPTOR.services_by_name['TimeService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\036com.daml.ledger.api.v1.testingB\025TimeServiceOuterClassZMgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1/testing\252\002\036Com.Daml.Ledger.Api.V1.Testing'
  _GETTIMEREQUEST._serialized_start=147
  _GETTIMEREQUEST._serialized_end=192
  _GETTIMERESPONSE._serialized_start=194
  _GETTIMERESPONSE._serialized_end=274
  _SETTIMEREQUEST._serialized_start=277
  _SETTIMEREQUEST._serialized_end=440
  _TIMESERVICE._serialized_start=443
  _TIMESERVICE._serialized_end=649
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
:mod:`dazl.ledger.blocking`
===========================

Thread-blocking protocols and base classes for connecting to a Daml ledger.

.. autoclass:: Connection()

.. autoclass:: QueryStream()
"""

import sys

from .. import Connection as _Connection, QueryStream as _QueryStream

if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable


__all__ = ["PackageService", "Connection", "QueryStream", "PackageLoader"]


class PackageService(Protocol):
    """
    Protocol that describe a service that provides package information. The :class:`Connection`
    protocol extends this interface.
    """

    async def get_package(self, __package_id) -> bytes:
        raise NotImplementedError

    async def list_package_ids(self):
        raise NotImplementedError


@runtime_checkable
class Connection(_Connection, Protocol):
    """
    Protocol that describes the interface for a blocking connection to a Daml ledger. The methods of
    an implementation should be thread-safe, with the exception of :meth:`open` and :meth:`close`.

    Since ``dazl`` activity is typically dominated by I/O, using a blocking :class:`Connection` with
    multiple threads is usually more expensive than using connections. You should consider
    using the asyncio variant of this API (:class:`dazl.ledger.aio.Connection`) instead.
    """

    def __enter__(self):
        raise NotImplementedError

    def __exit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError


@runtime_checkable
class QueryStream(_QueryStream, Protocol):
    """
    A blocking stream of offsets and events from a Daml ledger.
    """

    def __enter__(self):
        raise NotImplementedError

    def __exit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError

    def __iter__(self):
        raise NotImplementedError


# Imports internal to this package are at the bottom of the file to avoid circular dependencies
from .pkgloader import PackageLoader  # noqa



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from argparse import ArgumentParser
from asyncio import ensure_future, get_event_loop
import datetime
import json
import logging
from typing import Optional

from .. import LOG, Network
from ..client.config import NetworkConfig, configure_parser
from ..damlast.daml_types import con
from ..prim import Party
from ..values import Context, JsonDecoder
from ._base import CliCommand

__all__ = ["TailCommand"]


def elide_at(s: str, limit: "Optional[int]") -> str:
    if limit and len(s) > limit:
        return s[:limit] + "..."
    return s


def pprint_cdata(event, args):
    def default(o):
        if isinstance(o, (datetime.date, datetime.datetime)):
            return o.isoformat()

    format_args = {} if args.one_line else {"indent": 4}

    ctx = Context(JsonDecoder(), event.lookup)
    cdata = ctx.convert(con(event.cid.value_type), event.cdata)

    return elide_at(
        json.dumps(cdata, default=default, sort_keys=True, **format_args),
        None if args.details else 1024,
    )


def pprint_cid(event, args):
    tid = str(event.cid.value_type)

    delim_pos = tid.find(":")

    cid = elide_at(event.cid.value, None if args.details else 10)

    return f"{tid[delim_pos + 1:]}({cid})"


class TailCommand(CliCommand):
    name = "tail"

    def parser(self) -> ArgumentParser:
        arg_parser = ArgumentParser("dazl tail")

        configure_parser(arg_parser, config_file_support=True)
        arg_parser.add_argument(
            "--details",
            help="Ensure all output is printed with all details.",
            dest="details",
            action="store_true",
            default=False,
        )
        arg_parser.add_argument(
            "--one-line",
            help="Ensure all events are written as a single line",
            dest="one_line",
            action="store_true",
            default=False,
        )

        return arg_parser

    def execute(self, args) -> int:
        LOG.debug("Executing a tail...")

        loop = get_event_loop()
        loop.run_until_complete(self._main(args))

        return 0

    def setup_monitor_stream(self, network: "Network", party: "Party", args):
        party_client = network.aio_party(party)

        PLOG = logging.getLogger(f"tail-{party}")
        PLOG.setLevel(logging.INFO)

        @party_client.ledger_ready()
        async def on_ledger_ready(event):
            PLOG.info("...party log stream ready")

        @party_client.ledger_created("*")
        async def on_contract_create(event):
            PLOG.info(f"Create - {pprint_cid(event, args)} {pprint_cdata(event, args)}")

        @party_client.ledger_archived("*")
        async def on_contract_archived(event):
            PLOG.info(f"Archive - {pprint_cid(event, args)}")

    async def _main(self, args):
        final_config = NetworkConfig.get_config(args)

        network = Network()
        network.set_config(final_config)
        network.set_config(eager_package_fetch=False)

        for party in args.parties:
            self.setup_monitor_stream(network, party, args)

        await ensure_future(network.aio_run())



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module has been relocated to ``dazl.client``, ``dazl.damlast``, ``dazl.protocols``, or
``dazl.query``.
"""
from typing import TYPE_CHECKING, TypeVar, Union
import warnings

from ..client.errors import ConfigurationError, DazlPartyMissingError, UnknownTemplateWarning
from ..client.state import (
    ContractContextualData,
    ContractContextualDataCollection,
    ContractsHistoricalState,
    ContractsState,
)
from ..damlast.daml_lf_1 import TypeConName
from ..damlast.pkgfile import Dar
from ..prim import ContractData, ContractId as ContractId_, DazlError, DazlWarning, Party
from ..prim.errors import DazlImportError
from ..protocols.errors import ConnectionTimeoutError, UserTerminateRequest
from ..query import ContractMatch
from ..util.proc_util import ProcessDiedException

if TYPE_CHECKING:
    from .types import Type, TypeReference

T = TypeVar("T")


__all__ = [
    "ConfigurationError",
    "ConnectionTimeoutError",
    "ContractContextualData",
    "ContractContextualDataCollection",
    "ContractData",
    "ContractId",
    "ContractMatch",
    "ContractsHistoricalState",
    "ContractsState",
    "Dar",
    "DazlError",
    "DazlImportError",
    "DazlPartyMissingError",
    "DazlWarning",
    "Party",
    "ProcessDiedException",
    "UnknownTemplateWarning",
    "UserTerminateRequest",
]


class ContractId(ContractId_):
    __slots__ = ("_value_type_deprecated",)
    _value_type_deprecated: "TypeReference"

    def __init__(self, contract_id: str, template_id: "Union[str, Type, TypeConName]"):
        warnings.warn(
            "dazl.model.core.ContractId is deprecated; use dazl.prim.ContractId instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        from ..damlast.compat import parse_template

        if not isinstance(contract_id, str):
            raise ValueError("contract_id must be a string")

        value = contract_id
        value_type, value_type_deprecated = parse_template(template_id)

        super().__init__(value_type, value)
        object.__setattr__(self, "_value_type_deprecated", value_type_deprecated)

    @property
    def contract_id(self) -> str:
        """
        Get the raw contract ID value (for example, ``"#4:1"``).
        """
        warnings.warn(
            "ContractId.contract_id is deprecated; use ContractId.value instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.value

    @property
    def template_id(self) -> "TypeReference":
        """
        Get the type of template that is pointed to by this :class:`ContractId` as a
        :class:`TypeReference`. Note that usage of :class:`Type` and :class:`TypeReference` are
        deprecated, and :meth:`value_type` should be used instead.

        As of dazl 7.3.0, the :class:`TemplateId` is always normalized to a :class:`TypeReference`,
        regardless of what the :class:`ContractId` was constructed with.
        """
        warnings.warn(
            "ContractId.template_id is deprecated; use ContractId.value_type instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._value_type_deprecated

    def exercise(self, choice_name, arguments=None):
        """
        Create an :class:`ExerciseCommand` that represents the result of exercising a choice on this
        contract with the specified choice.

        :param choice_name:
            The name of the choice to exercise.
        :param arguments:
            (optional) A ``dict`` of named values to send as parameters to the choice exercise.
        """
        warnings.warn(
            "ContractId.exercise is deprecated; prefer calling dazl.ledger.Connection.exercise or "
            "dazl.client.PartyClient.submit_exercise, or use dazl.ledger.ExerciseCommand instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        from .writing import ExerciseCommand

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            return ExerciseCommand(self, choice_name, arguments=arguments)

    def replace(self, contract_id=None, template_id=None):
        """
        Return a new :class:`ContractId` instance replacing specified fields with values.
        """
        warnings.warn(
            "ContractId.replace is deprecated; simply construct a ContractId with the desired "
            "values instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            return ContractId(
                contract_id if contract_id is not None else self.value,
                template_id if template_id is not None else self.value_type,
            )

    def for_json(self):
        """
        Return the JSON representation of this contract. This is currently just the contract ID
        string itself.
        """
        return self.value


class CommandTimeoutError(DazlError):
    """
    Raised when a corresponding event for a command was not seen in the appropriate time window.
    """

    def __init__(self):
        warnings.warn(
            "This error is never raised; this symbol will be removed in dazl v9",
            DeprecationWarning,
            stacklevel=2,
        )


class ConnectionClosedError(DazlError):
    """
    Raised when trying to do something that requires a connection after connection pools have been
    closed.
    """

    def __init__(self):
        warnings.warn(
            "This error is never raised; this symbol will be removed in dazl v9",
            DeprecationWarning,
            stacklevel=2,
        )



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

__all__ = ["FrozenDict", "to_hashable"]


class FrozenDict(dict):
    """
    A special subclass of `dict` that is immutable and hashable. Instances of this "dict" can be
    used as keys in a Python dictionary.
    """

    def __hash__(self):
        return 0

    def __delitem__(self, *args, **kwargs):
        raise RuntimeError("frozendicts are immutable")

    def __setitem__(self, key, value):
        raise RuntimeError("frozendicts are immutable")

    def pop(self, *args, **kwargs):
        raise RuntimeError("frozendicts are immutable")

    def popitem(self, *args, **kwargs):
        raise RuntimeError("frozendicts are immutable")

    def setdefault(self, *args, **kwargs):
        raise RuntimeError("frozendicts are immutable")

    def update(self, *args, **kwargs):
        raise RuntimeError("frozendicts are immutable")

    def clear(self, *args, **kwargs):
        raise RuntimeError("frozendicts are immutable")


def to_hashable(obj):
    from collections import Collection, Mapping

    if isinstance(obj, Mapping):
        return FrozenDict(obj)
    elif isinstance(obj, str):
        return obj
    elif isinstance(obj, Collection):
        return tuple(obj)
    else:
        return obj



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
`asyncio`-flavored protocols and base classes for connecting to a Daml ledger.
"""

from collections import defaultdict
from inspect import iscoroutine
import sys
from typing import Any, Callable, TypeVar, Union, no_type_check
import warnings

from .. import Connection as _Connection, QueryStream as _QueryStream
from ...damlast.daml_lf_1 import TypeConName
from ...damlast.lookup import matching_normalizations, normalize
from ..api_types import ArchiveEvent, Boundary, CreateEvent, Event, ExerciseResponse
from ..errors import CallbackReturnWarning

if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable


__all__ = ["PackageService", "Connection", "QueryStream", "QueryStreamBase", "PackageLoader"]

Self = TypeVar("Self")
Ret = Union[None, CreateEvent, ExerciseResponse]
E = TypeVar("E", bound=Event)

CREATE_EVENT = "create"
ARCHIVE_EVENT = "archive"
BOUNDARY = "boundary"


class PackageService(Protocol):
    """
    Protocol that describe a service that provides package information. The :class:`Connection`
    protocol extends this interface.
    """

    async def get_package(self, __package_id) -> bytes:
        raise NotImplementedError

    async def list_package_ids(self):
        raise NotImplementedError


@runtime_checkable
class Connection(_Connection, Protocol):
    async def __aenter__(self):
        raise NotImplementedError

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        raise NotImplementedError


@runtime_checkable
class QueryStream(_QueryStream, Protocol):
    async def __aenter__(self):
        raise NotImplementedError

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError

    def __aiter__(self):
        raise NotImplementedError


class QueryStreamBase(QueryStream):
    """
    Base implementation of the :class:`QueryStream` protocol
    """

    @property
    def _callbacks(self):
        cb = getattr(self, "__cb", None)
        if cb is None:
            cb = defaultdict(list)
            object.__setattr__(self, "__cb", cb)
        return cb

    @no_type_check
    def on_boundary(self, *args):
        register(self, BOUNDARY, *args)

    @no_type_check
    def on_create(self, *args):
        register(self, CREATE_EVENT, *args)

    @no_type_check
    def on_archive(self, *args):
        register(self, ARCHIVE_EVENT, *args)

    async def __aenter__(self):
        """
        Prepare the stream.
        """
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Close the stream.
        """
        await self.close()

    async def close(self) -> None:
        """
        Close and dispose of any resources used by this stream.
        """

    async def run(self) -> None:
        """
        "Runs" the stream. This can be called as an alternative to :meth:`items` when using
        callback-based APIs.
        """
        async for _ in self:
            pass

    async def creates(self):
        """
        Return a stream of :class:`CreateEvent`s. This will include the contracts of the
        Active Contract Set, as well as create events over subsequent transactions.
        """
        async for item in self.items():
            if isinstance(item, CreateEvent):
                yield item

    async def events(self):
        """
        Return a stream of :class:`CreateEvent`s. This will include the contracts of the
        Active Contract Set, as well as create and archive events over subsequent transactions.
        """
        async for item in self.items():
            if isinstance(item, CreateEvent) or isinstance(item, ArchiveEvent):
                yield item

    def items(self):
        """
        Must be overridden by subclasses to provide a stream of events. The implementation is
        expected to call :meth:`_emit_create` and :meth:`_emit_archive` for every encountered event.
        """
        raise NotImplementedError

    def __aiter__(self):
        """
        Returns :meth:`items`, which includes all create and archive events, and boundaries.
        """
        return self.items()

    async def _emit(self, name: str, obj: "Any"):
        for cb in self._callbacks[name]:
            result = cb(obj)
            if result is not None and iscoroutine(result):
                result = await result

            # allow CreateEvent and ExerciseResponse as valid return types for callbacks because
            # it makes inline lambdas easier to write; we don't do anything with them
            if result is not None and not isinstance(result, (CreateEvent, ExerciseResponse)):
                warnings.warn(
                    "callbacks should not return anything; the result will be ignored",
                    CallbackReturnWarning,
                )

    async def _emit_create(self, event: "CreateEvent"):
        await self._emit(CREATE_EVENT, event)

    async def _emit_archive(self, event: "ArchiveEvent"):
        await self._emit(ARCHIVE_EVENT, event)

    async def _emit_boundary(self, event: "Boundary"):
        await self._emit(BOUNDARY, event)


def register(q: QueryStreamBase, name: str, *args):
    if len(args) == 0:
        return _register_decorator(q, name, None)
    elif len(args) == 1:
        if callable(args[0]):
            return _register(q, name, None, args[0])
        elif isinstance(args[0], (str, TypeConName)):
            return _register_decorator(q, name, args[0])
        else:
            raise ValueError("expected a template name or a callback here")
    elif len(args) == 2:
        template_id, fn = args
        _register(q, name, template_id, fn)
    else:
        raise ValueError("too many arguments")


def _register(
    q: QueryStreamBase,
    name: str,
    template_id: Union[None, str, TypeConName],
    fn: Callable[[E], None],
) -> Callable[[E], None]:
    template_search = normalize(template_id)

    def handler(event: E):
        if template_search in matching_normalizations(event.contract_id.value_type):
            fn(event)

    # noinspection PyProtectedMember
    q._callbacks[name].append(handler)
    return fn


def _register_decorator(q: QueryStreamBase, name: str, template_id: Union[None, str, TypeConName]):
    def decorator(fn):
        _register(q, name, template_id, fn)
        return fn

    return decorator


# Imports internal to this package are at the bottom of the file to avoid circular dependencies
from .pkgloader import PackageLoader  # noqa



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, sys, typing as _typing

from google.protobuf.internal.containers import RepeatedCompositeFieldContainer
from google.protobuf.message import Message as _Message

from .event_pb2 import CreatedEvent
from .transaction_filter_pb2 import TransactionFilter

if sys.version_info >= (3, 8):
    from typing import Literal as _L
else:
    from typing_extensions import Literal as _L

__all__ = [
    "GetActiveContractsRequest",
    "GetActiveContractsResponse",
]


class GetActiveContractsRequest(_Message):
    ledger_id: _builtins.str
    @property
    def filter(self) -> TransactionFilter: ...
    verbose: _builtins.bool
    def __init__(self, *, ledger_id: _typing.Optional[_builtins.str] = ..., filter: _typing.Optional[TransactionFilter] = ..., verbose: _typing.Optional[_builtins.bool] = ...): ...
    def HasField(self, field_name: _L["ledger_id", "filter", "verbose"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["ledger_id", "filter", "verbose"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class GetActiveContractsResponse(_Message):
    offset: _builtins.str
    workflow_id: _builtins.str
    @property
    def active_contracts(self) -> RepeatedCompositeFieldContainer[CreatedEvent]: ...
    def __init__(self, *, offset: _typing.Optional[_builtins.str] = ..., workflow_id: _typing.Optional[_builtins.str] = ..., active_contracts: _typing.Optional[_typing.Iterable[CreatedEvent]] = ...): ...
    def HasField(self, field_name: _L["offset", "workflow_id", "active_contracts"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["offset", "workflow_id", "active_contracts"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from .fmt_base import DEFAULT_FORMATTER_NAME
from .write import write_acs



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from enum import Enum
from typing import Dict, List, Tuple, Union

from google.protobuf.descriptor_pb2 import (
    DescriptorProto,
    EnumDescriptorProto,
    FieldDescriptorProto,
    FileDescriptorProto,
)

from .pb import py_message_package, py_scalar_type
from .types import (
    ANY,
    BOTTOM,
    COMPOSITE_CONTAINER,
    ITERABLE,
    MAP_CONTAINER,
    MAPPING,
    OPTIONAL,
    SCALAR_CONTAINER,
    PyType,
)

__all__ = ["Usage", "SymbolTable"]


class Usage(Enum):
    GETTER = 1
    ARG = 2
    ARG_STREAM = 3
    RET = 4
    RET_STREAM = 5
    RET_ASYNC = 6
    RET_STREAM_ASYNC = 7
    INIT = 8


class SymbolTable:
    """
    A map of "absolute" Protobuf imports to their Python types, taking into account that
    messages within the same file can be referenced without being fully-qualified.
    """

    def __init__(self):
        self._packages = {}  # type: Dict[str, str]
        self._imports = {}  # type: Dict[str, str]
        self._map_types = {}  # type: Dict[str, Tuple[FieldDescriptorProto, FieldDescriptorProto]]
        self._enums = {}  # type: Dict[str, Dict[str, int]]

    def load_file(self, fd: "FileDescriptorProto") -> None:
        """
        Load a single file.
        """
        prefix = f".{fd.package}"
        for ed in fd.enum_type:
            self.load_enum(ed, fd.name, prefix)
        for md in fd.message_type:
            self.load_message(md, fd.name, prefix, "")

    def load_enum(self, d: "EnumDescriptorProto", file_name: str, proto_prefix: str) -> None:
        """
        Load information about the specified enum.

        :param d:
            The :class:`EnumDescriptorProto` to load.
        :param file_name:
            The name of the file that contains this message.
        :param proto_prefix:
            The Protobuf namespace that contains this enum.
        """
        proto_type_name = proto_prefix + "." + d.name
        package = py_message_package(file_name)

        self._packages[proto_type_name] = package
        self._imports[proto_type_name] = d.name
        self._enums[proto_type_name] = {v.name: v.number for v in d.value}

    def load_message(
        self, d: "DescriptorProto", file_name: str, proto_prefix: str, local_prefix: str
    ) -> None:
        """
        Load information about the specified message.

        :param d:
            The :class:`EnumDescriptorProto` to load.
        :param file_name:
            The name of the file that contains this message.
        :param proto_prefix:
            The Protobuf namespace that contains this enum.
        :param local_prefix:
            The prefix of this type and any nested types, as it would be expressed in Python.
        """
        proto_type_name = proto_prefix + "." + d.name
        py_type_name = local_prefix + "." + d.name if local_prefix else d.name
        package = py_message_package(file_name)

        if d.options.map_entry:
            self._map_types[proto_type_name] = (d.field[0], d.field[1])
            return

        self._packages[proto_type_name] = package
        self._imports[proto_type_name] = py_type_name
        for ed in d.enum_type:
            self.load_enum(ed, file_name, proto_type_name)
        for nd in d.nested_type:
            self.load_message(nd, file_name, proto_type_name, py_type_name)

    def py_type(self, fd: "FieldDescriptorProto", usage: "Usage") -> "PyType":
        map_type = self._map_types.get(fd.type_name)
        if map_type is not None:
            key_type = self._base_py_type(map_type[0], Usage.GETTER)
            value_type = self._base_py_type(map_type[1], Usage.GETTER)
            if usage == Usage.GETTER:
                return MAP_CONTAINER.apply(key_type, value_type)
            elif usage == Usage.INIT:
                return OPTIONAL.apply(MAPPING.apply(key_type, value_type))
            elif usage == Usage.ARG or usage == Usage.RET:
                return MAPPING.apply(key_type, value_type)
            else:
                raise ValueError("unexpected map type in a stream context")

        py_type = self._base_py_type(fd, usage)
        if usage == Usage.GETTER:
            if fd.label == FieldDescriptorProto.LABEL_REPEATED:
                if py_type.settable:
                    return SCALAR_CONTAINER.apply(py_type)
                else:
                    return COMPOSITE_CONTAINER.apply(py_type)
            else:
                return py_type
        elif usage == Usage.INIT:
            if fd.label == FieldDescriptorProto.LABEL_REPEATED:
                return OPTIONAL.apply(ITERABLE.apply(py_type))
            elif fd.HasField("oneof_index"):
                return py_type
            else:
                return OPTIONAL.apply(py_type)
        elif usage == Usage.ARG:
            return py_type
        elif usage == Usage.ARG_STREAM:
            return ITERABLE.apply(py_type)
        elif usage == Usage.RET:
            return py_type
        elif usage == Usage.RET_STREAM:
            return PyType(py_str="_grpc.CallIterator").apply(py_type)
        elif usage == Usage.RET_ASYNC:
            return PyType(py_str="_grpc_aio.UnaryUnaryCall").apply(ANY, py_type)
        elif usage == Usage.RET_STREAM_ASYNC:
            return PyType(py_str="_grpc_aio.UnaryStreamCall").apply(ANY, py_type)
        else:
            raise ValueError

    def _base_py_type(self, fd: "FieldDescriptorProto", usage: "Usage") -> "PyType":
        """
        Return a Python string that represents the type of the requested field, ignoring whether
        or not it is a repeated field.
        """
        builtin_type = py_scalar_type(fd.type)
        if builtin_type is not None:
            return builtin_type

        enum_definition = self._enums.get(fd.type_name)
        if enum_definition is not None:
            # for enum definitions, the allowed constructors are literals over all the numeric AND
            # string representations for those types
            literals = []  # type: List[Union[str, int]]
            for key, value in enum_definition.items():
                if usage == Usage.INIT:
                    # for __init__ parameters for messages, the string values of enum entries are
                    # allowed; they are NOT allowed as setters for properties
                    literals.append(key)
                literals.append(value)

            if literals:
                # the Pythonic representation of a list is exactly what we need here
                return PyType(py_str=f"_L{literals}")
            else:
                # there are no allowable values; use the bottom type
                return BOTTOM

        type_str = self._imports.get(fd.type_name)
        if type_str is not None:
            python_package = self._packages[fd.type_name]
            return PyType(py_str=type_str, imports={python_package: [type_str.partition(".")[0]]})

        raise ValueError(f"could not compute a Pythonic representation of {fd.type_name!r}")



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/package_service.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n,com/daml/ledger/api/v1/package_service.proto\x12\x16\x63om.daml.ledger.api.v1\"2\n\x13ListPackagesRequest\x12\x1b\n\tledger_id\x18\x01 \x01(\tR\x08ledgerId\"7\n\x14ListPackagesResponse\x12\x1f\n\x0bpackage_ids\x18\x01 \x03(\tR\npackageIds\"O\n\x11GetPackageRequest\x12\x1b\n\tledger_id\x18\x01 \x01(\tR\x08ledgerId\x12\x1d\n\npackage_id\x18\x02 \x01(\tR\tpackageId\"\x9c\x01\n\x12GetPackageResponse\x12I\n\rhash_function\x18\x01 \x01(\x0e\x32$.com.daml.ledger.api.v1.HashFunctionR\x0chashFunction\x12\'\n\x0f\x61rchive_payload\x18\x02 \x01(\x0cR\x0e\x61rchivePayload\x12\x12\n\x04hash\x18\x03 \x01(\tR\x04hash\"U\n\x17GetPackageStatusRequest\x12\x1b\n\tledger_id\x18\x01 \x01(\tR\x08ledgerId\x12\x1d\n\npackage_id\x18\x02 \x01(\tR\tpackageId\"h\n\x18GetPackageStatusResponse\x12L\n\x0epackage_status\x18\x01 \x01(\x0e\x32%.com.daml.ledger.api.v1.PackageStatusR\rpackageStatus*,\n\rPackageStatus\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0e\n\nREGISTERED\x10\x01*\x1a\n\x0cHashFunction\x12\n\n\x06SHA256\x10\x00\x32\xd7\x02\n\x0ePackageService\x12i\n\x0cListPackages\x12+.com.daml.ledger.api.v1.ListPackagesRequest\x1a,.com.daml.ledger.api.v1.ListPackagesResponse\x12\x63\n\nGetPackage\x12).com.daml.ledger.api.v1.GetPackageRequest\x1a*.com.daml.ledger.api.v1.GetPackageResponse\x12u\n\x10GetPackageStatus\x12/.com.daml.ledger.api.v1.GetPackageStatusRequest\x1a\x30.com.daml.ledger.api.v1.GetPackageStatusResponseB\x92\x01\n\x16\x63om.daml.ledger.api.v1B\x18PackageServiceOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\xaa\x02\x16\x43om.Daml.Ledger.Api.V1b\x06proto3')

_PACKAGESTATUS = DESCRIPTOR.enum_types_by_name['PackageStatus']
PackageStatus = enum_type_wrapper.EnumTypeWrapper(_PACKAGESTATUS)
_HASHFUNCTION = DESCRIPTOR.enum_types_by_name['HashFunction']
HashFunction = enum_type_wrapper.EnumTypeWrapper(_HASHFUNCTION)
UNKNOWN = 0
REGISTERED = 1
SHA256 = 0


_LISTPACKAGESREQUEST = DESCRIPTOR.message_types_by_name['ListPackagesRequest']
_LISTPACKAGESRESPONSE = DESCRIPTOR.message_types_by_name['ListPackagesResponse']
_GETPACKAGEREQUEST = DESCRIPTOR.message_types_by_name['GetPackageRequest']
_GETPACKAGERESPONSE = DESCRIPTOR.message_types_by_name['GetPackageResponse']
_GETPACKAGESTATUSREQUEST = DESCRIPTOR.message_types_by_name['GetPackageStatusRequest']
_GETPACKAGESTATUSRESPONSE = DESCRIPTOR.message_types_by_name['GetPackageStatusResponse']
ListPackagesRequest = _reflection.GeneratedProtocolMessageType('ListPackagesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTPACKAGESREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.package_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.ListPackagesRequest)
  })
_sym_db.RegisterMessage(ListPackagesRequest)

ListPackagesResponse = _reflection.GeneratedProtocolMessageType('ListPackagesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTPACKAGESRESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.package_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.ListPackagesResponse)
  })
_sym_db.RegisterMessage(ListPackagesResponse)

GetPackageRequest = _reflection.GeneratedProtocolMessageType('GetPackageRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETPACKAGEREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.package_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.GetPackageRequest)
  })
_sym_db.RegisterMessage(GetPackageRequest)

GetPackageResponse = _reflection.GeneratedProtocolMessageType('GetPackageResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETPACKAGERESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.package_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.GetPackageResponse)
  })
_sym_db.RegisterMessage(GetPackageResponse)

GetPackageStatusRequest = _reflection.GeneratedProtocolMessageType('GetPackageStatusRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETPACKAGESTATUSREQUEST,
  '__module__' : 'com.daml.ledger.api.v1.package_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.GetPackageStatusRequest)
  })
_sym_db.RegisterMessage(GetPackageStatusRequest)

GetPackageStatusResponse = _reflection.GeneratedProtocolMessageType('GetPackageStatusResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETPACKAGESTATUSRESPONSE,
  '__module__' : 'com.daml.ledger.api.v1.package_service_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.GetPackageStatusResponse)
  })
_sym_db.RegisterMessage(GetPackageStatusResponse)

_PACKAGESERVICE = DESCRIPTOR.services_by_name['PackageService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026com.daml.ledger.api.v1B\030PackageServiceOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\252\002\026Com.Daml.Ledger.Api.V1'
  _PACKAGESTATUS._serialized_start=614
  _PACKAGESTATUS._serialized_end=658
  _HASHFUNCTION._serialized_start=660
  _HASHFUNCTION._serialized_end=686
  _LISTPACKAGESREQUEST._serialized_start=72
  _LISTPACKAGESREQUEST._serialized_end=122
  _LISTPACKAGESRESPONSE._serialized_start=124
  _LISTPACKAGESRESPONSE._serialized_end=179
  _GETPACKAGEREQUEST._serialized_start=181
  _GETPACKAGEREQUEST._serialized_end=260
  _GETPACKAGERESPONSE._serialized_start=263
  _GETPACKAGERESPONSE._serialized_end=419
  _GETPACKAGESTATUSREQUEST._serialized_start=421
  _GETPACKAGESTATUSREQUEST._serialized_end=506
  _GETPACKAGESTATUSRESPONSE._serialized_start=508
  _GETPACKAGESTATUSRESPONSE._serialized_end=612
  _PACKAGESERVICE._serialized_start=689
  _PACKAGESERVICE._serialized_end=1032
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, sys, typing as _typing

from google.protobuf.message import Message as _Message

from .commands_pb2 import Commands
from .transaction_pb2 import Transaction, TransactionTree

if sys.version_info >= (3, 8):
    from typing import Literal as _L
else:
    from typing_extensions import Literal as _L

__all__ = [
    "SubmitAndWaitRequest",
    "SubmitAndWaitForTransactionIdResponse",
    "SubmitAndWaitForTransactionResponse",
    "SubmitAndWaitForTransactionTreeResponse",
]


class SubmitAndWaitRequest(_Message):
    @property
    def commands(self) -> Commands: ...
    def __init__(self, *, commands: _typing.Optional[Commands] = ...): ...
    def HasField(self, field_name: _L["commands"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["commands"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class SubmitAndWaitForTransactionIdResponse(_Message):
    transaction_id: _builtins.str
    completion_offset: _builtins.str
    def __init__(self, *, transaction_id: _typing.Optional[_builtins.str] = ..., completion_offset: _typing.Optional[_builtins.str] = ...): ...
    def HasField(self, field_name: _L["transaction_id", "completion_offset"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["transaction_id", "completion_offset"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class SubmitAndWaitForTransactionResponse(_Message):
    @property
    def transaction(self) -> Transaction: ...
    completion_offset: _builtins.str
    def __init__(self, *, transaction: _typing.Optional[Transaction] = ..., completion_offset: _typing.Optional[_builtins.str] = ...): ...
    def HasField(self, field_name: _L["transaction", "completion_offset"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["transaction", "completion_offset"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class SubmitAndWaitForTransactionTreeResponse(_Message):
    @property
    def transaction(self) -> TransactionTree: ...
    completion_offset: _builtins.str
    def __init__(self, *, transaction: _typing.Optional[TransactionTree] = ..., completion_offset: _typing.Optional[_builtins.str] = ...): ...
    def HasField(self, field_name: _L["transaction", "completion_offset"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["transaction", "completion_offset"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Endpoints for managing parties/bots connected via a dazl client.
"""

from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING, Collection, NoReturn, Optional, Sequence
import warnings

from ..client import Bot, _NetworkImpl
from ..client.bots import SourceLocation
from ..prim import Party
from ..prim.errors import DazlImportError

if TYPE_CHECKING:
    from aiohttp import web


@dataclass(frozen=True)
class BotInfoList:
    bots: "Collection[BotInfo]"


@dataclass(frozen=True)
class BotInfo:
    id: str
    name: str
    party: "Optional[Party]"
    state: str
    entries: "Sequence[BotInfoEntry]"


@dataclass(frozen=True)
class BotInfoEntry:
    event_key: str
    source_location: "Optional[SourceLocation]"


def build_routes(network_impl: "_NetworkImpl") -> "Collection[web.AbstractRouteDef]":
    try:
        from aiohttp import web
    except ImportError:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            raise DazlImportError(
                "aiohttp", "server routes could not be built because aiohttp is not installed"
            )

    routes = web.RouteTableDef()

    @routes.get("/parties")
    async def get_parties(_: "web.Request") -> "web.Response":
        json_dict = dict(parties=[{"id": party} for party in network_impl.parties()])
        return web.json_response(json_dict)

    @routes.get("/bots")
    async def get_bots(_: "web.Request") -> "web.Response":
        bot_infos = [
            _get_bot_info(bot)
            for party_impl in network_impl.party_impls()
            for bot in party_impl.bots
        ]
        return web.json_response(asdict(BotInfoList(bots=bot_infos)))

    @routes.get("/bots/{bot_id}")
    async def get_bot_by_id(request: "web.Request") -> "web.Response":
        bot = get_bot(request)
        return web.json_response(asdict(_get_bot_info(bot)))

    @routes.post("/bots/{bot_id}/state/{action}")
    async def change_bot_state(request: "web.Request") -> "web.Response":
        bot = get_bot(request)
        action = request.match_info["action"]

        if action.lower() == "pause":
            bot.pause()
            return web.HTTPAccepted()
        elif action.lower() == "resume":
            bot.resume()
            return web.HTTPAccepted()
        else:
            return web.HTTPBadRequest()

    def get_bot(request: "web.Request") -> "Bot":
        """
        Return the bot referenced by the request. Throw HTTPNotFound if a bot with the ID specified
        in the URL does not exist.
        """
        bot_id = request.match_info["bot_id"]
        bot = network_impl.find_bot(bot_id)
        if bot is None:
            raise web.HTTPNotFound()
        return bot

    return routes


def _get_bot_info(bot: "Bot") -> "BotInfo":
    entries = [
        BotInfoEntry(event_key=entry.event_key, source_location=entry.source_location)
        for entry in bot.entries()
    ]

    return BotInfo(
        id=bot.id, party=bot.party, name=bot.name, state=bot.state.value, entries=entries
    )



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

from .value_pb2 import Enum, GenMap, Identifier, List, Map, Optional, Record, RecordField, Value, Variant
from .commands_pb2 import Command, Commands, CreateAndExerciseCommand, CreateCommand, ExerciseByKeyCommand, ExerciseCommand
from .completion_pb2 import Completion
from .event_pb2 import ArchivedEvent, CreatedEvent, Event, ExercisedEvent
from .experimental_features_pb2 import CommandDeduplicationFeatures, CommandDeduplicationPeriodSupport, CommandDeduplicationType, ExperimentalCommitterEventLog, ExperimentalContractIds, ExperimentalFeatures, ExperimentalOptionalLedgerId, ExperimentalSelfServiceErrorCodes, ExperimentalStaticTime
from .ledger_offset_pb2 import LedgerOffset
from .transaction_pb2 import Transaction, TransactionTree, TreeEvent
from .transaction_filter_pb2 import Filters, InclusiveFilters, TransactionFilter
from .active_contracts_service_pb2 import GetActiveContractsRequest, GetActiveContractsResponse
from .active_contracts_service_pb2_grpc import ActiveContractsServiceStub
from .command_completion_service_pb2 import Checkpoint, CompletionEndRequest, CompletionEndResponse, CompletionStreamRequest, CompletionStreamResponse
from .command_completion_service_pb2_grpc import CommandCompletionServiceStub
from .command_service_pb2 import SubmitAndWaitForTransactionIdResponse, SubmitAndWaitForTransactionResponse, SubmitAndWaitForTransactionTreeResponse, SubmitAndWaitRequest
from .command_service_pb2_grpc import CommandServiceStub
from .command_submission_service_pb2 import SubmitRequest
from .command_submission_service_pb2_grpc import CommandSubmissionServiceStub
from .ledger_configuration_service_pb2 import GetLedgerConfigurationRequest, GetLedgerConfigurationResponse, LedgerConfiguration
from .ledger_configuration_service_pb2_grpc import LedgerConfigurationServiceStub
from .ledger_identity_service_pb2 import GetLedgerIdentityRequest, GetLedgerIdentityResponse
from .ledger_identity_service_pb2_grpc import LedgerIdentityServiceStub
from .package_service_pb2 import GetPackageRequest, GetPackageResponse, GetPackageStatusRequest, GetPackageStatusResponse, HashFunction, ListPackagesRequest, ListPackagesResponse, PackageStatus
from .package_service_pb2_grpc import PackageServiceStub
from .transaction_service_pb2 import GetFlatTransactionResponse, GetLedgerEndRequest, GetLedgerEndResponse, GetTransactionByEventIdRequest, GetTransactionByIdRequest, GetTransactionResponse, GetTransactionTreesResponse, GetTransactionsRequest, GetTransactionsResponse
from .transaction_service_pb2_grpc import TransactionServiceStub
from .version_service_pb2 import FeaturesDescriptor, GetLedgerApiVersionRequest, GetLedgerApiVersionResponse, UserManagementFeature
from .version_service_pb2_grpc import VersionServiceStub

__all__ = [
    "ActiveContractsServiceStub",
    "ArchivedEvent",
    "Checkpoint",
    "Command",
    "CommandCompletionServiceStub",
    "CommandDeduplicationFeatures",
    "CommandDeduplicationPeriodSupport",
    "CommandDeduplicationType",
    "CommandServiceStub",
    "CommandSubmissionServiceStub",
    "Commands",
    "Completion",
    "CompletionEndRequest",
    "CompletionEndResponse",
    "CompletionStreamRequest",
    "CompletionStreamResponse",
    "CreateAndExerciseCommand",
    "CreateCommand",
    "CreatedEvent",
    "Enum",
    "Event",
    "ExerciseByKeyCommand",
    "ExerciseCommand",
    "ExercisedEvent",
    "ExperimentalCommitterEventLog",
    "ExperimentalContractIds",
    "ExperimentalFeatures",
    "ExperimentalOptionalLedgerId",
    "ExperimentalSelfServiceErrorCodes",
    "ExperimentalStaticTime",
    "FeaturesDescriptor",
    "Filters",
    "GenMap",
    "GetActiveContractsRequest",
    "GetActiveContractsResponse",
    "GetFlatTransactionResponse",
    "GetLedgerApiVersionRequest",
    "GetLedgerApiVersionResponse",
    "GetLedgerConfigurationRequest",
    "GetLedgerConfigurationResponse",
    "GetLedgerEndRequest",
    "GetLedgerEndResponse",
    "GetLedgerIdentityRequest",
    "GetLedgerIdentityResponse",
    "GetPackageRequest",
    "GetPackageResponse",
    "GetPackageStatusRequest",
    "GetPackageStatusResponse",
    "GetTransactionByEventIdRequest",
    "GetTransactionByIdRequest",
    "GetTransactionResponse",
    "GetTransactionTreesResponse",
    "GetTransactionsRequest",
    "GetTransactionsResponse",
    "HashFunction",
    "Identifier",
    "InclusiveFilters",
    "LedgerConfiguration",
    "LedgerConfigurationServiceStub",
    "LedgerIdentityServiceStub",
    "LedgerOffset",
    "List",
    "ListPackagesRequest",
    "ListPackagesResponse",
    "Map",
    "Optional",
    "PackageServiceStub",
    "PackageStatus",
    "Record",
    "RecordField",
    "SubmitAndWaitForTransactionIdResponse",
    "SubmitAndWaitForTransactionResponse",
    "SubmitAndWaitForTransactionTreeResponse",
    "SubmitAndWaitRequest",
    "SubmitRequest",
    "Transaction",
    "TransactionFilter",
    "TransactionServiceStub",
    "TransactionTree",
    "TreeEvent",
    "UserManagementFeature",
    "Value",
    "Variant",
    "VersionServiceStub",
]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'python',
 '_dazl_pb': 'python/_dazl_pb',
 '_dazl_pb.plugin_python': 'python/_dazl_pb/plugin_python',
 '_dazl_pb.protoc': 'python/_dazl_pb/protoc',
 '_dazl_pb.syntax': 'python/_dazl_pb/syntax',
 '_dazl_pb.syntax.python': 'python/_dazl_pb/syntax/python',
 '_dazl_pb.util': 'python/_dazl_pb/util'}

packages = \
['_dazl_pb',
 '_dazl_pb.plugin_python',
 '_dazl_pb.protoc',
 '_dazl_pb.syntax',
 '_dazl_pb.syntax.python',
 '_dazl_pb.util',
 'dazl',
 'dazl._gen',
 'dazl._gen.com',
 'dazl._gen.com.daml',
 'dazl._gen.com.daml.daml_lf_1_14',
 'dazl._gen.com.daml.ledger',
 'dazl._gen.com.daml.ledger.api',
 'dazl._gen.com.daml.ledger.api.v1',
 'dazl._gen.com.daml.ledger.api.v1.admin',
 'dazl._gen.com.daml.ledger.api.v1.testing',
 'dazl._repr',
 'dazl.cli',
 'dazl.client',
 'dazl.damlast',
 'dazl.ledger',
 'dazl.ledger.aio',
 'dazl.ledger.blocking',
 'dazl.ledger.config',
 'dazl.ledger.grpc',
 'dazl.ledgerutil',
 'dazl.metrics',
 'dazl.model',
 'dazl.pretty',
 'dazl.pretty.table',
 'dazl.prim',
 'dazl.protocols',
 'dazl.protocols.v0',
 'dazl.protocols.v1',
 'dazl.protocols.v1.model',
 'dazl.query',
 'dazl.scheduler',
 'dazl.server',
 'dazl.testing',
 'dazl.util',
 'dazl.values']

package_data = \
{'': ['*']}

install_requires = \
['googleapis_common_protos>=1,<2',
 'grpcio>=1.32.0',
 'protobuf>=3.19.0',
 'requests',
 'semver',
 'toposort']

extras_require = \
{':python_full_version >= "3.6.0" and python_full_version < "3.7.0"': ['async_exit_stack',
                                                                       'dataclasses'],
 'oauth': ['google-auth', 'oauthlib'],
 'prometheus': ['prometheus_client'],
 'pygments': ['pygments'],
 'server': ['aiohttp'],
 'tls-testing': ['pyOpenSSL']}

entry_points = \
{'console_scripts': ['dazl = dazl.cli:main',
                     'protoc-gen-dazl-python = _dazl_pb.plugin_python:main']}

setup_kwargs = {
    'name': 'dazl',
    'version': '7.9.0',
    'description': 'high-level Ledger API client for Daml ledgers',
    'long_description': 'dazl\n====\n\n[![License](https://img.shields.io/badge/License-Apache%202.0-blue.svg)](https://github.com/digital-asset/dazl-client/blob/main/LICENSE)\n<a href="https://circleci.com/gh/digital-asset/dazl-client">\n<img src="https://circleci.com/gh/digital-asset/dazl-client.svg?style=svg">\n</a>\n\nCopyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All Rights Reserved.\nSPDX-License-Identifier: Apache-2.0\n\n\nRich Python bindings for accessing Ledger API-based applications.\n\nDocumentation\n-------------\nThe user documentation is available online [here](https://digital-asset.github.io/dazl-client).\n\nInstallation\n------------\nIf you just want to use the library, you can install it locally with `pip`:\n```sh\npip install --user dazl\n```\n\nRequirements\n------------\n* Python 3.6+\n* [Daml Connect](https://www.daml.com)\n* Python gRPC libraries (1.32.0 or later) and Protobuf\n\n**WARNING:** The next major version of dazl (v8.0.0) will require **Python 3.7** or later.\n\nExamples\n--------\n\nAll of the examples below assume you imported `dazl`, and are running a ledger with the default scenario generated with `daml new`.\n\nConnect to the ledger and submit a single command:\n\n```py\nimport asyncio\nimport dazl\n\nasync def main():\n    async with dazl.connect(url=\'http://localhost:6865\', act_as=\'Alice\') as client:\n        contract = { \'issuer\' : \'Alice\', \'owner\' : \'Alice\', \'name\' : \'hello world!\' }\n        await client.create(\'Main:Asset\', contract)\n\n# Python 3.7+\nasyncio.run(main())\n\n# Python 3.6+\nloop = asyncio.get_event_loop()\nloop.run_until_complete(main())\n```\n\nConnect to the ledger as a single party, print all contracts, and close:\n\n```py\nimport asyncio\nimport dazl\nfrom dazl.ledgerutil import ACS\n\nasync def main():\n    async with dazl.connect(url=\'http://localhost:6865\', read_as=\'Alice\') as conn:\n        async with ACS(conn, {"*": {}}) as acs:\n            snapshot = await acs.read()\n\n    print(snapshot)\n\n# Python 3.7+\nasyncio.run(main())\n\n# Python 3.6+\nloop = asyncio.get_event_loop()\nloop.run_until_complete(main())\n```\n\nBuilding locally\n----------------\n\nYou will need additional dependencies to build locally:\n\n* GNU Make 4.3 or later\n* [Poetry](https://python-poetry.org/) for build/dependency management\n\nOnce you have these prerequisites in place:\n\n```sh\nmake build\n```\n\nIf you see errors about incompatible python versions, switch your environment to python3 using `poetry env use python3`, for instance.\n\nTests\n-----\n\nTests in dazl are written using [pytest](https://docs.pytest.org/en/latest/). You can run them by doing:\n\n```sh\nmake test\n```\n\nSupport\n-------\n\nThe dazl library is supported by the Daml community. If you are in need of support, have questions or just want to engage in friendly conversation anything Daml, contact us on our [Daml Community Forum](https://discuss.daml.com).\n',
    'author': 'Davin K. Tanabe',
    'author_email': 'davin.tanabe@digitalasset.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/digital-asset/dazl-client',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/event.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import value_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_value__pb2
from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"com/daml/ledger/api/v1/event.proto\x12\x16\x63om.daml.ledger.api.v1\x1a\"com/daml/ledger/api/v1/value.proto\x1a\x1egoogle/protobuf/wrappers.proto\"\xa8\x01\n\x05\x45vent\x12@\n\x07\x63reated\x18\x01 \x01(\x0b\x32$.com.daml.ledger.api.v1.CreatedEventH\x00R\x07\x63reated\x12\x43\n\x08\x61rchived\x18\x03 \x01(\x0b\x32%.com.daml.ledger.api.v1.ArchivedEventH\x00R\x08\x61rchivedB\x07\n\x05\x65ventJ\x04\x08\x02\x10\x03R\texercised\"\xca\x03\n\x0c\x43reatedEvent\x12\x19\n\x08\x65vent_id\x18\x01 \x01(\tR\x07\x65ventId\x12\x1f\n\x0b\x63ontract_id\x18\x02 \x01(\tR\ncontractId\x12\x43\n\x0btemplate_id\x18\x03 \x01(\x0b\x32\".com.daml.ledger.api.v1.IdentifierR\ntemplateId\x12@\n\x0c\x63ontract_key\x18\x07 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.ValueR\x0b\x63ontractKey\x12I\n\x10\x63reate_arguments\x18\x04 \x01(\x0b\x32\x1e.com.daml.ledger.api.v1.RecordR\x0f\x63reateArguments\x12\'\n\x0fwitness_parties\x18\x05 \x03(\tR\x0ewitnessParties\x12 \n\x0bsignatories\x18\x08 \x03(\tR\x0bsignatories\x12\x1c\n\tobservers\x18\t \x03(\tR\tobservers\x12\x43\n\x0e\x61greement_text\x18\x06 \x01(\x0b\x32\x1c.google.protobuf.StringValueR\ragreementText\"\xb9\x01\n\rArchivedEvent\x12\x19\n\x08\x65vent_id\x18\x01 \x01(\tR\x07\x65ventId\x12\x1f\n\x0b\x63ontract_id\x18\x02 \x01(\tR\ncontractId\x12\x43\n\x0btemplate_id\x18\x03 \x01(\x0b\x32\".com.daml.ledger.api.v1.IdentifierR\ntemplateId\x12\'\n\x0fwitness_parties\x18\x04 \x03(\tR\x0ewitnessParties\"\xdb\x03\n\x0e\x45xercisedEvent\x12\x19\n\x08\x65vent_id\x18\x01 \x01(\tR\x07\x65ventId\x12\x1f\n\x0b\x63ontract_id\x18\x02 \x01(\tR\ncontractId\x12\x43\n\x0btemplate_id\x18\x03 \x01(\x0b\x32\".com.daml.ledger.api.v1.IdentifierR\ntemplateId\x12\x16\n\x06\x63hoice\x18\x05 \x01(\tR\x06\x63hoice\x12\x46\n\x0f\x63hoice_argument\x18\x06 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.ValueR\x0e\x63hoiceArgument\x12%\n\x0e\x61\x63ting_parties\x18\x07 \x03(\tR\ractingParties\x12\x1c\n\tconsuming\x18\x08 \x01(\x08R\tconsuming\x12\'\n\x0fwitness_parties\x18\n \x03(\tR\x0ewitnessParties\x12&\n\x0f\x63hild_event_ids\x18\x0b \x03(\tR\rchildEventIds\x12\x46\n\x0f\x65xercise_result\x18\x0c \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.ValueR\x0e\x65xerciseResultJ\x04\x08\x04\x10\x05J\x04\x08\t\x10\nB\x89\x01\n\x16\x63om.daml.ledger.api.v1B\x0f\x45ventOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\xaa\x02\x16\x43om.Daml.Ledger.Api.V1b\x06proto3')



_EVENT = DESCRIPTOR.message_types_by_name['Event']
_CREATEDEVENT = DESCRIPTOR.message_types_by_name['CreatedEvent']
_ARCHIVEDEVENT = DESCRIPTOR.message_types_by_name['ArchivedEvent']
_EXERCISEDEVENT = DESCRIPTOR.message_types_by_name['ExercisedEvent']
Event = _reflection.GeneratedProtocolMessageType('Event', (_message.Message,), {
  'DESCRIPTOR' : _EVENT,
  '__module__' : 'com.daml.ledger.api.v1.event_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Event)
  })
_sym_db.RegisterMessage(Event)

CreatedEvent = _reflection.GeneratedProtocolMessageType('CreatedEvent', (_message.Message,), {
  'DESCRIPTOR' : _CREATEDEVENT,
  '__module__' : 'com.daml.ledger.api.v1.event_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.CreatedEvent)
  })
_sym_db.RegisterMessage(CreatedEvent)

ArchivedEvent = _reflection.GeneratedProtocolMessageType('ArchivedEvent', (_message.Message,), {
  'DESCRIPTOR' : _ARCHIVEDEVENT,
  '__module__' : 'com.daml.ledger.api.v1.event_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.ArchivedEvent)
  })
_sym_db.RegisterMessage(ArchivedEvent)

ExercisedEvent = _reflection.GeneratedProtocolMessageType('ExercisedEvent', (_message.Message,), {
  'DESCRIPTOR' : _EXERCISEDEVENT,
  '__module__' : 'com.daml.ledger.api.v1.event_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.ExercisedEvent)
  })
_sym_db.RegisterMessage(ExercisedEvent)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026com.daml.ledger.api.v1B\017EventOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\252\002\026Com.Daml.Ledger.Api.V1'
  _EVENT._serialized_start=131
  _EVENT._serialized_end=299
  _CREATEDEVENT._serialized_start=302
  _CREATEDEVENT._serialized_end=760
  _ARCHIVEDEVENT._serialized_start=763
  _ARCHIVEDEVENT._serialized_end=948
  _EXERCISEDEVENT._serialized_start=951
  _EXERCISEDEVENT._serialized_end=1426
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from types import MappingProxyType
from typing import Collection, Mapping, Optional

from .builtins import builtins
from .daml_lf_1 import Block, Expr, Type, ValName
from .protocols import SymbolLookup
from .util import package_local_name
from .visitor import IdentityVisitor


class RewriteVisitor(IdentityVisitor):
    def __init__(
        self,
        lookup: "SymbolLookup",
        always_expand: bool = False,
        val_blacklist: "Optional[Collection[ValName]]" = None,
        expr_args: "Optional[Mapping[str, Expr]]" = None,
        type_args: "Optional[Mapping[str, Type]]" = None,
    ):
        self.lookup = lookup
        self.always_expand = always_expand
        self.val_blacklist = frozenset(val_blacklist) if val_blacklist is not None else ()
        self.expr_args = MappingProxyType(expr_args or {})  # type: Mapping[str, Expr]
        self.type_args = MappingProxyType(type_args or {})  # type: Mapping[str, Type]

    def with_expr_var(self, var: str, value: "Expr") -> "RewriteVisitor":
        return type(self)(
            self.lookup,
            self.always_expand,
            self.val_blacklist,
            {**self.expr_args, var: value},
            self.type_args,
        )

    def with_expr_vars(self, var_values: "Mapping[str, Expr]") -> "RewriteVisitor":
        return type(self)(
            self.lookup,
            self.always_expand,
            self.val_blacklist,
            {**self.expr_args, **var_values},
            self.type_args,
        )

    def with_type_vars(self, type_values: "Mapping[str, Type]") -> "RewriteVisitor":
        return type(self)(
            self.lookup,
            self.always_expand,
            self.val_blacklist,
            self.expr_args,
            {**self.type_args, **type_values},
        )

    def without_val(self, val: "ValName") -> "RewriteVisitor":
        return type(self)(
            self.lookup,
            self.always_expand,
            {*self.val_blacklist, val},
            self.expr_args,
            self.type_args,
        )

    def resolve_val(self, val: "ValName") -> "Optional[Expr]":
        return self.lookup.value(val).expr if val not in self.val_blacklist else None

    def resolve_type_arg(self, type_arg: str) -> "Optional[Type]":
        return self.type_args.get(type_arg)

    def visit_expr_var(self, var: str) -> "Expr":
        expr = self.expr_args.get(var)
        return expr if expr is not None else Expr(var=var)

    def visit_type_var(self, var: "Type.Var") -> "Type":
        base_type = self.resolve_type_arg(var.var)
        if var.args:
            raise ValueError("higher kinded types are not yet currently supported")
        return base_type if base_type is not None else Type(var=var)


class ExpandVisitor(RewriteVisitor):
    def visit_expr_val(self, val: "ValName") -> "Expr":
        builtin = builtins.resolve(val)
        if builtin is not None:
            return Expr(val=val)
        if (
            self.always_expand
            or val._name[0][0] == "$"
            or package_local_name(val) == "DA.Internal.Template:toParties"
        ):
            val_expr = self.resolve_val(val)
            if val_expr is not None:
                child = self.without_val(val)
                return child.visit_expr(val_expr)
            else:
                print(
                    f"Failed to resolve {package_local_name(val)} against blacklist {self.val_blacklist}"
                )

        return super(ExpandVisitor, self).visit_expr_val(val)

    def visit_expr_ty_app(self, ty_app: "Expr.TyApp") -> "Expr":
        if ty_app.expr.ty_abs is not None:
            # perform type substitutions on the body, and remove the outer layers
            fn_vars = {
                p.var: self.visit_type(body)
                for p, body in zip(ty_app.expr.ty_abs.param, ty_app.types)
            }
            unapplied_args = ty_app.types[len(ty_app.expr.ty_abs.param) :]
            remaining_fn_args = ty_app.expr.ty_abs.param[len(ty_app.types) :]
            scope = self.with_type_vars(fn_vars)
            eval_body = scope.visit_expr(ty_app.expr.ty_abs.body)

            if unapplied_args:
                # the arguments are overapplied on the body of the function; substitute the passed
                # parameters in the body of the function, and the remaining parameters are applied
                # to the result of evaluation of the body
                return Expr(ty_app=Expr.TyApp(expr=eval_body, types=unapplied_args))
            elif remaining_fn_args:
                # the arguments are underapplied on the body of the function; substitute what we
                # can and return a partially-applied function
                return Expr(ty_abs=Expr.TyAbs(body=eval_body, param=remaining_fn_args))
            else:
                return eval_body
        elif ty_app.expr.builtin is not None:
            # this is a type abstraction over a built-in function; this TyApp serves more as
            # adornment over the bulit-in function tha anything else
            return Expr(ty_app=ty_app)
        else:
            return super(ExpandVisitor, self).visit_expr_ty_app(ty_app)

    #    ty_app.types[0]
    #     # some ty_apps get automatically expanded to compact representations that avoid a lot of the
    #     # declared machinery
    #     if ty_app.expr.val is not None and ty_app.expr.val.full_name_unambiguous == 'DA.Internal.Prelude:concat':
    #         # return a simple function call that takes two lists and concatenates them
    #         var_type = list_type(list_type(ty_app.types[0]))
    #         abs_param = (VarWithType('xxs', var_type),)  # type: Sequence[VarWithType]
    #         return Expr(abs=Expr.Abs(param=abs_param, body=Expr(builtin=BuiltinFunction.CONCAT_LIST)))
    #     else:
    #         return super(ExpandVisitor, self).visit_expr_ty_app(ty_app)
    #     # ty_app.types
    #     # ty_app.expr

    # def visit_expr_ty_abs(self, ty_abs: 'Expr.TyAbs') -> 'Expr':
    #     self.with_type_vars(ty_abs.param[0].var)
    #     ty_abs.param[0].var
    #     ty_abs.body
    #
    # def visit_expr_case(self, case: 'Case') -> 'Expr':
    #     case.scrut

    def visit_expr_let(self, let: "Block") -> "Expr":
        new_body = let.body
        for binding in reversed(let.bindings):
            scope = self.with_expr_var(binding.binder.var, binding.bound)
            new_body = scope.visit_expr(new_body)
        return new_body

    def visit_type_forall(self, forall: "Type.Forall") -> "Type":
        new_vars = [tvwk for tvwk in forall.vars if tvwk.var not in self.type_args]
        new_body = self.visit_type(forall.body)
        return Type(forall=Type.Forall(vars=new_vars, body=new_body))


class SimplifyVisitor(RewriteVisitor):
    def visit_expr_app(self, app: "Expr.App"):
        # first, if this is a built-in function, immediately apply its simplification logic on its
        # evaluated arguments
        if app.fun.ty_app is not None and app.fun.ty_app.expr.val is not None:
            from .builtins import builtins

            builtin = builtins.resolve(app.fun.ty_app.expr.val)
            if builtin is not None:
                type_args = app.fun.ty_app.types
                new_args = [self.visit_expr(arg) for arg in app.args]
                result = builtin.simplify(type_args, new_args)
                if result is not None:
                    return result

        # evaluate the function body (which we'll again re-evaluate with the arguments in a moment)
        new_fun = self.visit_expr(app.fun)

        if new_fun.abs is not None:
            #     builtin = builtins.resolve(new_fun.abs.body.ty_app.expr.val)
            #     if builtin is not None:
            #         type_args = new_fun.ty_app.types
            #         new_args = [self.visit_expr(arg) for arg in app.args]
            #         return builtin.simplify(type_args, new_args)

            # if new_fun.abs.val is not None and new_fun.abs.val in builtins:
            #     new_args = [self.visit_expr(arg) for arg in app.args]
            #     builtin = builtins.resolve(new_fun.abs.val)
            #     return builtin.simplify(None, new_args)

            if new_fun.abs.body.var is not None:
                # the body is a single variable; if it matches the name of the argument, then this
                # is the identity function and we simply return the argument
                if len(new_fun.abs.param) == 1 and len(app.args) == 1:
                    if new_fun.abs.param[0].var == new_fun.abs.body.var:
                        return self.visit_expr(app.args[0])

            if new_fun.abs.body.prim_lit is not None or new_fun.abs.body.prim_con is not None:
                # The body is actually NOT a function of its parameters, so we can skip evaluation of
                # the relevant parameters. This happens often enough in typical DAML models that it
                # is very much worth shortcutting this case
                if len(new_fun.abs.param) == 1 and len(app.args) == 1:
                    # just return the literal value and that's it
                    return new_fun.abs.body

            new_args = [self.visit_expr(arg) for arg in app.args]
            fn_vars = {p.var: self.visit_expr(body) for p, body in zip(new_fun.abs.param, new_args)}
            unapplied_args = new_args[len(new_fun.abs.param) :]
            remaining_fn_args = new_fun.abs.param[len(new_args) :]
            scope = self.with_expr_vars(fn_vars)
            eval_body = scope.visit_expr(new_fun.abs.body)

            if unapplied_args:
                # the arguments are overapplied on the body of the function; substitute the passed
                # parameters in the body of the function, and the remaining parameters are applied
                # to the result of evaluation of the body
                return self.visit_expr(Expr(app=Expr.App(fun=eval_body, args=unapplied_args)))
            elif remaining_fn_args:
                # the arguments are underapplied on the body of the function; substitute what we
                # can and return a partially-applied function
                return Expr(abs=Expr.Abs(body=eval_body, param=remaining_fn_args))
            else:
                return eval_body
        # elif app.fun.builtin is not None or (app.fun.ty_app is not None and app.fun.ty_app.expr.builtin is not None):
        #     builtin = app.fun.builtin or app.fun.ty_app.expr.builtin
        #     if builtin == BuiltinFunction.FOLDR:
        #         # foldr(func, acc, xs)
        #
        else:
            return super(SimplifyVisitor, self).visit_expr_app(app)

    def visit_expr_cons(self, cons: "Expr.Cons"):
        new_fronts = [self.visit_expr(front) for front in cons.front]
        new_tail = self.visit_expr(cons.tail)
        if new_tail.cons:
            new_fronts.extend(new_tail.cons.front)
            return Expr(
                cons=Expr.Cons(
                    type=cons.type,
                    front=tuple(new_fronts),
                    tail=Expr(nil=Expr.Nil(type=new_tail.cons.type)),
                )
            )
        else:
            return Expr(cons=Expr.Cons(type=cons.type, front=tuple(new_fronts), tail=new_tail))

    def visit_expr_rec_proj(self, rec_proj: "Expr.RecProj"):
        new_record = self.visit_expr(rec_proj.record)
        if new_record.rec_con is not None:
            for fwt in new_record.rec_con.fields:
                if fwt.field == rec_proj.field:
                    return self.visit_expr(fwt.expr)
            raise RuntimeError("rec_proj over a rec_con that did not have the necessary field")
        else:
            return Expr(
                rec_proj=Expr.RecProj(tycon=rec_proj.tycon, field=rec_proj.field, record=new_record)
            )

    def visit_expr_struct_proj(self, struct_proj: "Expr.StructProj"):
        new_tuple = self.visit_expr(struct_proj.struct)
        if new_tuple.struct_con is not None:
            for fwt in new_tuple.struct_con.fields:
                if fwt.field == struct_proj.field:
                    return self.visit_expr(fwt.expr)
            raise RuntimeError("rec_proj over a rec_con that did not have the necessary field")
        else:
            return Expr(struct_proj=Expr.StructProj(field=struct_proj.field, struct=new_tuple))



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, sys, typing as _typing

from google.protobuf.descriptor import EnumDescriptor
from google.protobuf.internal.containers import RepeatedScalarFieldContainer
from google.protobuf.message import Message as _Message

if sys.version_info >= (3, 8):
    from typing import Literal as _L
else:
    from typing_extensions import Literal as _L

__all__ = [
    "ListPackagesRequest",
    "ListPackagesResponse",
    "GetPackageRequest",
    "GetPackageResponse",
    "GetPackageStatusRequest",
    "GetPackageStatusResponse",
]

class PackageStatus:
    DESCRIPTOR: _typing.ClassVar[EnumDescriptor] = ...
    UNKNOWN: _typing.ClassVar[_L[0]] = ...
    REGISTERED: _typing.ClassVar[_L[1]] = ...
UNKNOWN = _L[0]
REGISTERED = _L[1]

class HashFunction:
    DESCRIPTOR: _typing.ClassVar[EnumDescriptor] = ...
    SHA256: _typing.ClassVar[_L[0]] = ...
SHA256 = _L[0]


class ListPackagesRequest(_Message):
    ledger_id: _builtins.str
    def __init__(self, *, ledger_id: _typing.Optional[_builtins.str] = ...): ...
    def HasField(self, field_name: _L["ledger_id"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["ledger_id"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class ListPackagesResponse(_Message):
    @property
    def package_ids(self) -> RepeatedScalarFieldContainer[_builtins.str]: ...
    def __init__(self, *, package_ids: _typing.Optional[_typing.Iterable[_builtins.str]] = ...): ...
    def HasField(self, field_name: _L["package_ids"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["package_ids"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class GetPackageRequest(_Message):
    ledger_id: _builtins.str
    package_id: _builtins.str
    def __init__(self, *, ledger_id: _typing.Optional[_builtins.str] = ..., package_id: _typing.Optional[_builtins.str] = ...): ...
    def HasField(self, field_name: _L["ledger_id", "package_id"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["ledger_id", "package_id"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class GetPackageResponse(_Message):
    @property
    def hash_function(self) -> _L[0]: ...
    archive_payload: _builtins.bytes
    hash: _builtins.str
    def __init__(self, *, hash_function: _typing.Optional[_L['SHA256', 0]] = ..., archive_payload: _typing.Optional[_builtins.bytes] = ..., hash: _typing.Optional[_builtins.str] = ...): ...
    def HasField(self, field_name: _L["hash_function", "archive_payload", "hash"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["hash_function", "archive_payload", "hash"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class GetPackageStatusRequest(_Message):
    ledger_id: _builtins.str
    package_id: _builtins.str
    def __init__(self, *, ledger_id: _typing.Optional[_builtins.str] = ..., package_id: _typing.Optional[_builtins.str] = ...): ...
    def HasField(self, field_name: _L["ledger_id", "package_id"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["ledger_id", "package_id"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class GetPackageStatusResponse(_Message):
    @property
    def package_status(self) -> _L[0, 1]: ...
    def __init__(self, *, package_status: _typing.Optional[_L['UNKNOWN', 0, 'REGISTERED', 1]] = ...): ...
    def HasField(self, field_name: _L["package_status"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["package_status"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file


__all__ = []  # type: ignore



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, sys, typing as _typing

from google.protobuf.message import Message as _Message

if sys.version_info >= (3, 8):
    from typing import Literal as _L
else:
    from typing_extensions import Literal as _L

__all__ = [
    "ResetRequest",
]


class ResetRequest(_Message):
    ledger_id: _builtins.str
    def __init__(self, *, ledger_id: _typing.Optional[_builtins.str] = ...): ...
    def HasField(self, field_name: _L["ledger_id"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["ledger_id"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from ._sandbox import SandboxLauncher
from .connect import connect_with_new_party

__all__ = ["connect_with_new_party", "sandbox", "SandboxLauncher"]

sandbox = SandboxLauncher



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from io import StringIO
import os
from typing import TextIO

from google.protobuf.compiler.plugin_pb2 import CodeGeneratorRequest, CodeGeneratorResponse
from google.protobuf.descriptor_pb2 import FileDescriptorProto, ServiceDescriptorProto

from .. import util
from ..syntax.python import ImportContext, SymbolTable, Usage, all_decl, py_service_package
from ._header import HEADER

__all__ = ["run_plugin"]


def run_plugin(request: "CodeGeneratorRequest") -> "CodeGeneratorResponse":
    """
    Generate .pyi (typing stubs files) for gPRC services.
    """
    # build a mapping of all types that we'll ever need to generate any of the requested files
    symbol_table = SymbolTable()
    for fd in request.proto_file:
        symbol_table.load_file(fd)

    request = util.services_only(request)

    return CodeGeneratorResponse(
        file=[
            typing_file(pf, ImportContext(symbol_table))
            for pf in request.proto_file
            if pf.name in request.file_to_generate
        ]
    )


def typing_file(fd: "FileDescriptorProto", ictx: "ImportContext") -> "CodeGeneratorResponse.File":
    name = os.path.splitext(fd.name)[0] + "_pb2_grpc.pyi"
    with StringIO() as buf:
        for sd in fd.service:
            write_service(buf, sd, ictx)

        body = buf.getvalue()

    ictx.add_system_import("", "builtins as _builtins")
    ictx.add_system_import("", "typing as _typing")
    ictx.add_import("", "grpc as _grpc")
    ictx.add_import("grpc", "aio as _grpc_aio")

    imports = ictx.py_import_block(py_service_package(fd.name))

    all_str = all_decl(md.name + "Stub" for md in fd.service)

    return CodeGeneratorResponse.File(name=name, content=f"{HEADER}\n{imports}\n{all_str}\n{body}")


def write_service(buf: "TextIO", sd: "ServiceDescriptorProto", ictx: "ImportContext") -> None:
    buf.write("\n")

    with StringIO() as stub_buf, StringIO() as abuf, StringIO() as bbuf:
        astub = f"_{sd.name}AsyncStub"
        bstub = f"_{sd.name}BlockingStub"

        stub_buf.write(
            f"# noinspection PyPep8Naming,DuplicatedCode\n"
            f"class {sd.name}Stub:\n"
            f"    @classmethod\n"
            f"    @_typing.overload\n"
            f"    def __new__(cls, channel: _grpc.Channel) -> {bstub}: ...  # type: ignore\n"
            f"    @classmethod\n"
            f"    @_typing.overload\n"
            f"    def __new__(cls, channel: _grpc_aio.Channel) -> {astub}: ...  # type: ignore\n"
        )

        abuf.write("# noinspection PyPep8Naming,DuplicatedCode\n")
        abuf.write(f"class {astub}({sd.name}Stub):\n")

        bbuf.write("# noinspection PyPep8Naming,DuplicatedCode\n")
        bbuf.write(f"class {bstub}({sd.name}Stub):\n")

        for method in sd.method:
            arg = ictx.py_type(
                method.input_type, Usage.ARG_STREAM if method.client_streaming else Usage.ARG
            )
            a_ret = ictx.py_type(
                method.output_type,
                Usage.RET_STREAM_ASYNC if method.server_streaming else Usage.RET_ASYNC,
            )
            b_ret = ictx.py_type(
                method.output_type, Usage.RET_STREAM if method.server_streaming else Usage.RET
            )

            timeout_param = "timeout: _typing.Optional[float] = ..."
            credentials_param = "credentials: _typing.Optional[_grpc.CallCredentials] = ..."
            wait_for_ready_param = "wait_for_ready: _typing.Optional[bool] = ..."
            compression_param = "compression: _typing.Optional[_grpc.Compression] = ..."

            stub_buf.write(f"    def {method.name}(self, __1: {arg.py_str}, *, ")
            stub_buf.write(timeout_param)
            stub_buf.write(
                ", metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ...,"
            )
            stub_buf.write(
                f" {credentials_param}, {wait_for_ready_param}, {compression_param}) -> "
            )
            stub_buf.write(f"_typing.Union[{b_ret.py_str}, {a_ret.py_str}]: ...\n")

            # when using AsyncIO-flavored channels, the optional parameters are explicitly
            # keyword-only
            abuf.write(f"    def {method.name}(self, __1: {arg.py_str}, *, ")
            abuf.write(timeout_param)
            abuf.write(", metadata: _typing.Optional[_grpc_aio.Metadata] = ...,")
            abuf.write(f" {credentials_param}, {wait_for_ready_param}, {compression_param}) -> ")
            abuf.write(a_ret.py_str)
            abuf.write(": ...\n")

            bbuf.write(
                f"    def {method.name}(self, __1: {arg.py_str}, {timeout_param}, "
                + "metadata: _typing.Optional[_typing.Tuple[_typing.Tuple[str, _typing.Union[str, bytes]], ...]] = ..."
                + f", {credentials_param}, {wait_for_ready_param}, {compression_param}) -> "
                + b_ret.py_str
                + ": ...\n"
            )

        buf.write(stub_buf.getvalue() + "\n")
        buf.write(bbuf.getvalue() + "\n")
        buf.write(abuf.getvalue())
